// ISO 15924 Codes for the representation of names of scripts.
// License: Public Domain.
// This file is autogenerated.
// Based on data from: https://unicode.org/iso15924/
// file: iso15924-codes.html
// url: https://unicode.org/iso15924/iso15924-codes.html

#ifndef ISO15924_DATA_H
#define ISO15924_DATA_H

#ifdef __cplusplus
extern "C" {
#endif

#define ISO15924_DATA_SZ	210

struct iso15924_rec {
	/**
	 * Code.
	 */
	const char* code;
	/**
	 * English name.
	 */
	const char* name;
	/**
	 * Alias.
	 */
	const char* alias;
	/**
	 * Numeric value.
	 */
	unsigned int num;
};

extern const struct iso15924_rec iso15924_data[];

#ifdef __cplusplus
}
#endif

#endif // ISO15924_DATA_H
