// ISO 3166-1 Country Codes.
// License: Public Domain.
// This file is autogenerated.
// Based on data from: https://www.iso.org/iso-3166-country-codes.html

#ifndef ISO3166_1_DATA_H
#define ISO3166_1_DATA_H

#ifdef __cplusplus
extern "C" {
#endif

#define ISO3166_1_DATA_SZ	251

struct iso3166_1_rec {
	/**
	 * Alpha-3 code.
	 */
	const char* alpha3;
	/**
	 * Alpha-2 code.
	 */
	const char* alpha2;
	/**
	 * English short name.
	 */
	const char* name;
	/**
	 * Numeric value.
	 */
	unsigned int num;
};

extern const struct iso3166_1_rec iso3166_1_data[];

#ifdef __cplusplus
}
#endif

#endif // ISO3166_1_DATA_H
