
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int pl_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00d3,
	0x00f3,
	0xf0f0ffff, 0x0104, 0x0105,	// range
	0xf0f0ffff, 0x0106, 0x0107,	// range
	0xf0f0ffff, 0x0118, 0x0119,	// range
	0xf0f0ffff, 0x0141, 0x0142,	// range
	0xf0f0ffff, 0x0143, 0x0144,	// range
	0xf0f0ffff, 0x015a, 0x015b,	// range
	0xf0f0ffff, 0x0179, 0x017a,	// range
	0xf0f0ffff, 0x017b, 0x017c,	// range
};
#define PL_LANG_ORTH_SZ	32
