
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int sk_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c1,
	0x00c4,
	0x00c9,
	0x00cd,
	0x00d3,
	0x00d4,
	0x00da,
	0x00dd,
	0x00e1,
	0x00e4,
	0x00e9,
	0x00ed,
	0x00f3,
	0x00f4,
	0x00fa,
	0x00fd,
	0xf0f0ffff, 0x010c, 0x010d,	// range
	0xf0f0ffff, 0x010e, 0x010f,	// range
	0xf0f0ffff, 0x0139, 0x013a,	// range
	0xf0f0ffff, 0x013d, 0x013e,	// range
	0xf0f0ffff, 0x0147, 0x0148,	// range
	0xf0f0ffff, 0x0154, 0x0155,	// range
	0xf0f0ffff, 0x0160, 0x0161,	// range
	0xf0f0ffff, 0x0164, 0x0165,	// range
	0xf0f0ffff, 0x017d, 0x017e,	// range
};
#define SK_LANG_ORTH_SZ	49
