
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int hsb_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00d3,
	0x00f3,
	0xf0f0ffff, 0x0106, 0x0107,	// range
	0xf0f0ffff, 0x010c, 0x010d,	// range
	0xf0f0ffff, 0x011a, 0x011b,	// range
	0xf0f0ffff, 0x0141, 0x0144,	// range
	0xf0f0ffff, 0x0158, 0x0159,	// range
	0xf0f0ffff, 0x0160, 0x0161,	// range
	0xf0f0ffff, 0x0179, 0x017a,	// range
	0xf0f0ffff, 0x017d, 0x017e,	// range
};
#define HSB_LANG_ORTH_SZ	32
