// ISO 3166-1 Country Codes.
// License: Public Domain.
// This file is autogenerated.
// Based on data from: https://www.iso.org/iso-3166-country-codes.html

#include "iso-3166-1_data.h"

#include <stddef.h>

const struct iso3166_1_rec iso3166_1_data[] = {
	{"AFG", "AF", "Afghanistan", 4},
	{"ALB", "AL", "Albania", 8},
	{"DZA", "DZ", "Algeria", 12},
	{"ASM", "AS", "American Samoa", 16},
	{"AND", "AD", "Andorra", 20},
	{"AGO", "AO", "Angola", 24},
	{"AIA", "AI", "Anguilla", 660},
	{"ATA", "AQ", "Antarctica", 10},
	{"ATG", "AG", "Antigua and Barbuda", 28},
	{"ARG", "AR", "Argentina", 32},
	{"ARM", "AM", "Armenia", 51},
	{"ABW", "AW", "Aruba", 533},
	{"AUS", "AU", "Australia", 36},
	{"AUT", "AT", "Austria", 40},
	{"AZE", "AZ", "Azerbaijan", 31},
	{"BHS", "BS", "Bahamas (the)", 44},
	{"BHR", "BH", "Bahrain", 48},
	{"BGD", "BD", "Bangladesh", 50},
	{"BRB", "BB", "Barbados", 52},
	{"BLR", "BY", "Belarus", 112},
	{"BEL", "BE", "Belgium", 56},
	{"BLZ", "BZ", "Belize", 84},
	{"BEN", "BJ", "Benin", 204},
	{"BMU", "BM", "Bermuda", 60},
	{"BTN", "BT", "Bhutan", 64},
	{"BOL", "BO", "Bolivia (Plurinational State of)", 68},
	{"BES", "BQ", "Bonaire, Sint Eustatius and Saba", 535},
	{"BIH", "BA", "Bosnia and Herzegovina", 70},
	{"BWA", "BW", "Botswana", 72},
	{"BVT", "BV", "Bouvet Island", 74},
	{"BRA", "BR", "Brazil", 76},
	{"IOT", "IO", "British Indian Ocean Territory (the)", 86},
	{"BRN", "BN", "Brunei Darussalam", 96},
	{"BGR", "BG", "Bulgaria", 100},
	{"BFA", "BF", "Burkina Faso", 854},
	{"BDI", "BI", "Burundi", 108},
	{"CPV", "CV", "Cabo Verde", 132},
	{"KHM", "KH", "Cambodia", 116},
	{"CMR", "CM", "Cameroon", 120},
	{"CAN", "CA", "Canada", 124},
	{"CYM", "KY", "Cayman Islands (the)", 136},
	{"CAF", "CF", "Central African Republic (the)", 140},
	{"TCD", "TD", "Chad", 148},
	{"CHL", "CL", "Chile", 152},
	{"CHN", "CN", "China", 156},
	{"CXR", "CX", "Christmas Island", 162},
	{"CCK", "CC", "Cocos (Keeling) Islands (the)", 166},
	{"COL", "CO", "Colombia", 170},
	{"COM", "KM", "Comoros (the)", 174},
	{"COD", "CD", "Congo (the Democratic Republic of the)", 180},
	{"COG", "CG", "Congo (the)", 178},
	{"COK", "CK", "Cook Islands (the)", 184},
	{"CRI", "CR", "Costa Rica", 188},
	{"HRV", "HR", "Croatia", 191},
	{"CUB", "CU", "Cuba", 192},
	{"CUW", "CW", "Curaçao", 531},
	{"CYP", "CY", "Cyprus", 196},
	{"CZE", "CZ", "Czechia", 203},
	{"CIV", "CI", "Côte d'Ivoire", 384},
	{"DNK", "DK", "Denmark", 208},
	{"DJI", "DJ", "Djibouti", 262},
	{"DMA", "DM", "Dominica", 212},
	{"DOM", "DO", "Dominican Republic (the)", 214},
	{"ECU", "EC", "Ecuador", 218},
	{"EGY", "EG", "Egypt", 818},
	{"SLV", "SV", "El Salvador", 222},
	{"GNQ", "GQ", "Equatorial Guinea", 226},
	{"ERI", "ER", "Eritrea", 232},
	{"EST", "EE", "Estonia", 233},
	{"SWZ", "SZ", "Eswatini", 748},
	{"ETH", "ET", "Ethiopia", 231},
	{"FLK", "FK", "Falkland Islands (the) [Malvinas]", 238},
	{"FRO", "FO", "Faroe Islands (the)", 234},
	{"FJI", "FJ", "Fiji", 242},
	{"FIN", "FI", "Finland", 246},
	{"FRA", "FR", "France", 250},
	{"GUF", "GF", "French Guiana", 254},
	{"PYF", "PF", "French Polynesia", 258},
	{"ATF", "TF", "French Southern Territories (the)", 260},
	{"GAB", "GA", "Gabon", 266},
	{"GMB", "GM", "Gambia (the)", 270},
	{"GEO", "GE", "Georgia", 268},
	{"DEU", "DE", "Germany", 276},
	{"GHA", "GH", "Ghana", 288},
	{"GIB", "GI", "Gibraltar", 292},
	{"GRC", "GR", "Greece", 300},
	{"GRL", "GL", "Greenland", 304},
	{"GRD", "GD", "Grenada", 308},
	{"GLP", "GP", "Guadeloupe", 312},
	{"GUM", "GU", "Guam", 316},
	{"GTM", "GT", "Guatemala", 320},
	{"GGY", "GG", "Guernsey", 831},
	{"GIN", "GN", "Guinea", 324},
	{"GNB", "GW", "Guinea-Bissau", 624},
	{"GUY", "GY", "Guyana", 328},
	{"HTI", "HT", "Haiti", 332},
	{"HMD", "HM", "Heard Island and McDonald Islands", 334},
	{"VAT", "VA", "Holy See (the)", 336},
	{"HND", "HN", "Honduras", 340},
	{"HKG", "HK", "Hong Kong", 344},
	{"HUN", "HU", "Hungary", 348},
	{"ISL", "IS", "Iceland", 352},
	{"IND", "IN", "India", 356},
	{"IDN", "ID", "Indonesia", 360},
	{"IRN", "IR", "Iran (Islamic Republic of)", 364},
	{"IRQ", "IQ", "Iraq", 368},
	{"IRL", "IE", "Ireland", 372},
	{"IMN", "IM", "Isle of Man", 833},
	{"ISR", "IL", "Israel", 376},
	{"ITA", "IT", "Italy", 380},
	{"JAM", "JM", "Jamaica", 388},
	{"JPN", "JP", "Japan", 392},
	{"JEY", "JE", "Jersey", 832},
	{"JOR", "JO", "Jordan", 400},
	{"KAZ", "KZ", "Kazakhstan", 398},
	{"KEN", "KE", "Kenya", 404},
	{"KIR", "KI", "Kiribati", 296},
	{"PRK", "KP", "Korea (the Democratic People's Republic of)", 408},
	{"KOR", "KR", "Korea (the Republic of)", 410},
	{"KWT", "KW", "Kuwait", 414},
	{"KGZ", "KG", "Kyrgyzstan", 417},
	{"LAO", "LA", "Lao People's Democratic Republic (the)", 418},
	{"LVA", "LV", "Latvia", 428},
	{"LBN", "LB", "Lebanon", 422},
	{"LSO", "LS", "Lesotho", 426},
	{"LBR", "LR", "Liberia", 430},
	{"LBY", "LY", "Libya", 434},
	{"LIE", "LI", "Liechtenstein", 438},
	{"LTU", "LT", "Lithuania", 440},
	{"LUX", "LU", "Luxembourg", 442},
	{"MAC", "MO", "Macao", 446},
	{"MDG", "MG", "Madagascar", 450},
	{"MWI", "MW", "Malawi", 454},
	{"MYS", "MY", "Malaysia", 458},
	{"MDV", "MV", "Maldives", 462},
	{"MLI", "ML", "Mali", 466},
	{"MLT", "MT", "Malta", 470},
	{"MHL", "MH", "Marshall Islands (the)", 584},
	{"MTQ", "MQ", "Martinique", 474},
	{"MRT", "MR", "Mauritania", 478},
	{"MUS", "MU", "Mauritius", 480},
	{"MYT", "YT", "Mayotte", 175},
	{"MEX", "MX", "Mexico", 484},
	{"FSM", "FM", "Micronesia (Federated States of)", 583},
	{"MDA", "MD", "Moldova (the Republic of)", 498},
	{"MCO", "MC", "Monaco", 492},
	{"MNG", "MN", "Mongolia", 496},
	{"MNE", "ME", "Montenegro", 499},
	{"MSR", "MS", "Montserrat", 500},
	{"MAR", "MA", "Morocco", 504},
	{"MOZ", "MZ", "Mozambique", 508},
	{"MMR", "MM", "Myanmar", 104},
	{"NAM", "NA", "Namibia", 516},
	{"NRU", "NR", "Nauru", 520},
	{"NPL", "NP", "Nepal", 524},
	{"NLD", "NL", "Netherlands (the)", 528},
	{"NCL", "NC", "New Caledonia", 540},
	{"NZL", "NZ", "New Zealand", 554},
	{"NIC", "NI", "Nicaragua", 558},
	{"NER", "NE", "Niger (the)", 562},
	{"NGA", "NG", "Nigeria", 566},
	{"NIU", "NU", "Niue", 570},
	{"NFK", "NF", "Norfolk Island", 574},
	{"MKD", "MK", "North Macedonia", 807},
	{"MNP", "MP", "Northern Mariana Islands (the)", 580},
	{"NOR", "NO", "Norway", 578},
	{"OMN", "OM", "Oman", 512},
	{"PAK", "PK", "Pakistan", 586},
	{"PLW", "PW", "Palau", 585},
	{"PSE", "PS", "Palestine, State of", 275},
	{"PAN", "PA", "Panama", 591},
	{"PNG", "PG", "Papua New Guinea", 598},
	{"PRY", "PY", "Paraguay", 600},
	{"PER", "PE", "Peru", 604},
	{"PHL", "PH", "Philippines (the)", 608},
	{"PCN", "PN", "Pitcairn", 612},
	{"POL", "PL", "Poland", 616},
	{"PRT", "PT", "Portugal", 620},
	{"PRI", "PR", "Puerto Rico", 630},
	{"QAT", "QA", "Qatar", 634},
	{"ROU", "RO", "Romania", 642},
	{"RUS", "RU", "Russian Federation (the)", 643},
	{"RWA", "RW", "Rwanda", 646},
	{"REU", "RE", "Réunion", 638},
	{"BLM", "BL", "Saint Barthélemy", 652},
	{"SHN", "SH", "Saint Helena, Ascension and Tristan da Cunha", 654},
	{"KNA", "KN", "Saint Kitts and Nevis", 659},
	{"LCA", "LC", "Saint Lucia", 662},
	{"MAF", "MF", "Saint Martin (French part)", 663},
	{"SPM", "PM", "Saint Pierre and Miquelon", 666},
	{"VCT", "VC", "Saint Vincent and the Grenadines", 670},
	{"WSM", "WS", "Samoa", 882},
	{"SMR", "SM", "San Marino", 674},
	{"STP", "ST", "Sao Tome and Principe", 678},
	{"SAU", "SA", "Saudi Arabia", 682},
	{"SEN", "SN", "Senegal", 686},
	{"SRB", "RS", "Serbia", 688},
	{"SYC", "SC", "Seychelles", 690},
	{"SLE", "SL", "Sierra Leone", 694},
	{"SGP", "SG", "Singapore", 702},
	{"SXM", "SX", "Sint Maarten (Dutch part)", 534},
	{"SVK", "SK", "Slovakia", 703},
	{"SVN", "SI", "Slovenia", 705},
	{"SLB", "SB", "Solomon Islands", 90},
	{"SOM", "SO", "Somalia", 706},
	{"ZAF", "ZA", "South Africa", 710},
	{"SGS", "GS", "South Georgia and the South Sandwich Islands", 239},
	{"SSD", "SS", "South Sudan", 728},
	{"ESP", "ES", "Spain", 724},
	{"LKA", "LK", "Sri Lanka", 144},
	{"SDN", "SD", "Sudan (the)", 729},
	{"SUR", "SR", "Suriname", 740},
	{"SJM", "SJ", "Svalbard and Jan Mayen", 744},
	{"SWE", "SE", "Sweden", 752},
	{"CHE", "CH", "Switzerland", 756},
	{"SYR", "SY", "Syrian Arab Republic (the)", 760},
	{"TWN", "TW", "Taiwan (Province of China)", 158},
	{"TJK", "TJ", "Tajikistan", 762},
	{"TZA", "TZ", "Tanzania, the United Republic of", 834},
	{"THA", "TH", "Thailand", 764},
	{"TLS", "TL", "Timor-Leste", 626},
	{"TGO", "TG", "Togo", 768},
	{"TKL", "TK", "Tokelau", 772},
	{"TON", "TO", "Tonga", 776},
	{"TTO", "TT", "Trinidad and Tobago", 780},
	{"TUN", "TN", "Tunisia", 788},
	{"TUR", "TR", "Turkey", 792},
	{"TKM", "TM", "Turkmenistan", 795},
	{"TCA", "TC", "Turks and Caicos Islands (the)", 796},
	{"TUV", "TV", "Tuvalu", 798},
	{"UGA", "UG", "Uganda", 800},
	{"UKR", "UA", "Ukraine", 804},
	{"ARE", "AE", "United Arab Emirates (the)", 784},
	{"GBR", "GB", "United Kingdom of Great Britain and Northern Ireland (the)", 826},
	{"UMI", "UM", "United States Minor Outlying Islands (the)", 581},
	{"USA", "US", "United States of America (the)", 840},
	{"URY", "UY", "Uruguay", 858},
	{"UZB", "UZ", "Uzbekistan", 860},
	{"VUT", "VU", "Vanuatu", 548},
	{"VEN", "VE", "Venezuela (Bolivarian Republic of)", 862},
	{"VNM", "VN", "Viet Nam", 704},
	{"VGB", "VG", "Virgin Islands (British)", 92},
	{"VIR", "VI", "Virgin Islands (U.S.)", 850},
	{"WLF", "WF", "Wallis and Futuna", 876},
	{"ESH", "EH", "Western Sahara*", 732},
	{"YEM", "YE", "Yemen", 887},
	{"ZMB", "ZM", "Zambia", 894},
	{"ZWE", "ZW", "Zimbabwe", 716},
	{"ALA", "AX", "Åland Islands", 248},
	{"ANT", "AN", "Netherlands Antilles", 530},
	{"CSG", "CS", "Serbia and Montenegro", 891},
};

