
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ba_lang_orth_chars[] = {
	0xf0f0ffff, 0x0410, 0x044f,	// range
	0x0492,
	0x0493,
	0x0498,
	0x0499,
	0x04a0,
	0x04a1,
	0x04a2,
	0x04a3,
	0x04aa,
	0x04ab,
	0x04ae,
	0x04af,
	0x04ba,
	0x04bb,
	0x04d8,
	0x04d9,
	0x04e8,
	0x04e9,
};
#define BA_LANG_ORTH_SZ	21
