DOCX_TAG(abstractNum)
DOCX_TAG(abstractNumId)
DOCX_TAG(anchor)
DOCX_TAG(document)
DOCX_TAG(body)
DOCX_TAG(br)
DOCX_TAG(blip)
DOCX_TAG(blipFill)
DOCX_TAG(footnote)
DOCX_TAG(footnotes)
DOCX_TAG(footnoteReference)
DOCX_TAG(footnoteRef)
DOCX_TAG(endnote)
DOCX_TAG(endnotes)
DOCX_TAG(endnoteReference)
DOCX_TAG(endnoteRef)
DOCX_TAG(hyperlink)
DOCX_TAG(bookmarkStart)
DOCX_TAG(bookmarkEnd)
DOCX_TAG(drawing)
DOCX_TAG(fldChar)
DOCX_TAG(graphic)
DOCX_TAG(graphicData)
DOCX_TAG(gridCol)
DOCX_TAG(gridSpan)
DOCX_TAG(numPr)
DOCX_TAG(ilvl)
DOCX_TAG(inline)
DOCX_TAG(instrText)
DOCX_TAG(isLgl)
DOCX_TAG(lang)
DOCX_TAG(lvl)
DOCX_TAG(lvlOverride)
DOCX_TAG(lvlJc)
DOCX_TAG(lvlRestart)
DOCX_TAG(lvlText)
DOCX_TAG(num)
DOCX_TAG(numFmt)
DOCX_TAG(numbering)
DOCX_TAG(nvPicPr)
DOCX_TAG(outlineLvl)
DOCX_TAG(numId)
DOCX_TAG(pic)
DOCX_TAG(pPr)
DOCX_TAG(pPrDefault)
DOCX_TAG(rPr)
DOCX_TAG(spPr)
DOCX_TAG(start)
DOCX_TAG(rPrDefault)
DOCX_TAG(tab)
DOCX_TAG(tbl)
DOCX_TAG(tblGrid)
DOCX_TAG(tblPr)
DOCX_TAG(tc)
DOCX_TAG(tcPr)
DOCX_TAG(tr)
DOCX_TAG(trPr)
DOCX_TAG(b)
DOCX_TAG(i)
DOCX_TAG(u)
DOCX_TAG(strike)
DOCX_TAG(color)
DOCX_TAG(p)
DOCX_TAG(r)
DOCX_TAG(t)
DOCX_TAG(jc)
DOCX_TAG(rFonts)
DOCX_TAG(vertAlign)
DOCX_TAG(sz)
DOCX_TAG(vanish)
DOCX_TAG(pStyle)
DOCX_TAG(rStyle)
DOCX_TAG(ind)
DOCX_TAG(spacing)
DOCX_TAG(suff)
DOCX_TAG(textAlignment)
DOCX_TAG(suppressAutoHyphens)
DOCX_TAG(style)
DOCX_TAG(styles)
DOCX_TAG(docDefaults)
DOCX_TAG(name)
DOCX_TAG(basedOn)
DOCX_TAG(pageBreakBefore)
DOCX_TAG(keepNext)
DOCX_TAG(mirrorIndents)
DOCX_TAG(vMerge)
DOCX_NUM_FMT(aiueo)
DOCX_NUM_FMT(aiueoFullWidth)
DOCX_NUM_FMT(arabicAbjad)
DOCX_NUM_FMT(arabicAlpha)
DOCX_NUM_FMT(bullet)
DOCX_NUM_FMT(cardinalText)
DOCX_NUM_FMT(chicago)
DOCX_NUM_FMT(chineseCounting)
DOCX_NUM_FMT(chineseCountingThousand)
DOCX_NUM_FMT(chineseLegalSimplified)
DOCX_NUM_FMT(chosung)
DOCX_NUM_FMT(decimal)
DOCX_NUM_FMT(decimalEnclosedCircle)
DOCX_NUM_FMT(decimalEnclosedCircleChinese)
DOCX_NUM_FMT(decimalEnclosedFullstop)
DOCX_NUM_FMT(decimalEnclosedParen)
DOCX_NUM_FMT(decimalFullWidth)
DOCX_NUM_FMT(decimalFullWidth2)
DOCX_NUM_FMT(decimalHalfWidth)
DOCX_NUM_FMT(decimalZero)
DOCX_NUM_FMT(ganada)
DOCX_NUM_FMT(hebrew1)
DOCX_NUM_FMT(hebrew2)
DOCX_NUM_FMT(hex)
DOCX_NUM_FMT(hindiConsonants)
DOCX_NUM_FMT(hindiCounting)
DOCX_NUM_FMT(hindiNumbers)
DOCX_NUM_FMT(hindiVowels)
DOCX_NUM_FMT(ideographDigital)
DOCX_NUM_FMT(ideographEnclosedCircle)
DOCX_NUM_FMT(ideographLegalTraditional)
DOCX_NUM_FMT(ideographTraditional)
DOCX_NUM_FMT(ideographZodiac)
DOCX_NUM_FMT(ideographZodiacTraditional)
DOCX_NUM_FMT(iroha)
DOCX_NUM_FMT(irohaFullWidth)
DOCX_NUM_FMT(japaneseCounting)
DOCX_NUM_FMT(japaneseDigitalTenThousand)
DOCX_NUM_FMT(japaneseLegal)
DOCX_NUM_FMT(koreanCounting)
DOCX_NUM_FMT(koreanDigital)
DOCX_NUM_FMT(koreanDigital2)
DOCX_NUM_FMT(koreanLegal)
DOCX_NUM_FMT(lowerLetter)
DOCX_NUM_FMT(lowerRoman)
DOCX_NUM_FMT(none)
DOCX_NUM_FMT(numberInDash)
DOCX_NUM_FMT(ordinal)
DOCX_NUM_FMT(ordinalText)
DOCX_NUM_FMT(russianLower)
DOCX_NUM_FMT(russianUpper)
DOCX_NUM_FMT(taiwaneseCounting)
DOCX_NUM_FMT(taiwaneseCountingThousand)
DOCX_NUM_FMT(taiwaneseDigital)
DOCX_NUM_FMT(thaiCounting)
DOCX_NUM_FMT(thaiLetters)
DOCX_NUM_FMT(thaiNumbers)
DOCX_NUM_FMT(upperLetter)
DOCX_NUM_FMT(upperRoman)
DOCX_NUM_FMT(vietnameseCounting)
#undef DOCX_NUM_FMT
#undef DOCX_TAG
