
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int lt_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x0104, 0x0105,	// range
	0xf0f0ffff, 0x010c, 0x010d,	// range
	0xf0f0ffff, 0x0116, 0x0117,	// range
	0xf0f0ffff, 0x0118, 0x0119,	// range
	0xf0f0ffff, 0x012e, 0x012f,	// range
	0xf0f0ffff, 0x0160, 0x0161,	// range
	0xf0f0ffff, 0x016a, 0x016b,	// range
	0xf0f0ffff, 0x0172, 0x0173,	// range
	0xf0f0ffff, 0x017d, 0x017e,	// range
};
#define LT_LANG_ORTH_SZ	33
