
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int iu_lang_orth_chars[] = {
	0xf0f0ffff, 0x1401, 0x1406,	// range
	0xf0f0ffff, 0x140a, 0x140b,	// range
	0xf0f0ffff, 0x142f, 0x1434,	// range
	0xf0f0ffff, 0x1438, 0x1439,	// range
	0x1449,
	0xf0f0ffff, 0x144c, 0x1451,	// range
	0xf0f0ffff, 0x1455, 0x1456,	// range
	0x1466,
	0xf0f0ffff, 0x146b, 0x1470,	// range
	0xf0f0ffff, 0x1472, 0x1473,	// range
	0x1483,
	0xf0f0ffff, 0x1489, 0x148e,	// range
	0xf0f0ffff, 0x1490, 0x1491,	// range
	0x14a1,
	0xf0f0ffff, 0x14a3, 0x14a8,	// range
	0xf0f0ffff, 0x14aa, 0x14ab,	// range
	0x14bb,
	0xf0f0ffff, 0x14c0, 0x14c5,	// range
	0xf0f0ffff, 0x14c7, 0x14c8,	// range
	0x14d0,
	0xf0f0ffff, 0x14d3, 0x14d8,	// range
	0xf0f0ffff, 0x14da, 0x14db,	// range
	0x14ea,
	0xf0f0ffff, 0x14ed, 0x14f2,	// range
	0xf0f0ffff, 0x14f4, 0x14f5,	// range
	0x14fa,
	0x14fc,
	0x14fe,
	0x1500,
	0x1502,
	0x1505,
	0xf0f0ffff, 0x1526, 0x152b,	// range
	0xf0f0ffff, 0x152d, 0x152e,	// range
	0x153e,
	0x1542,
	0xf0f0ffff, 0x1545, 0x1549,	// range
	0xf0f0ffff, 0x154b, 0x154c,	// range
	0x1550,
	0xf0f0ffff, 0x1553, 0x155a,	// range
	0x155d,
	0xf0f0ffff, 0x1575, 0x1577,	// range
	0xf0f0ffff, 0x1579, 0x157c,	// range
	0xf0f0ffff, 0x157e, 0x1585,	// range
	0xf0f0ffff, 0x158b, 0x1596,	// range
	0xf0f0ffff, 0x15a0, 0x15a6,	// range
	0xf0f0ffff, 0x15a8, 0x15ae,	// range
	0x166f,
	0xf0f0ffff, 0x1670, 0x1676,	// range
};
#define IU_LANG_ORTH_SZ	108
