
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int fat_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c3,
	0x00e3,
	0x00d1,
	0x00d5,
	0x00f1,
	0x00f5,
	0x0128,
	0x0129,
	0x0168,
	0x0169,
	0x0186,
	0x0254,
	0x0190,
	0x025b,
	0x0303,
	0x0306,
	0x0329,
	0x1ebc,
	0x1ebd,
	0x1ef8,
	0x1ef9,
};
#define FAT_LANG_ORTH_SZ	27
