
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int bo_lang_orth_chars[] = {
	0xf0f0ffff, 0x0f40, 0x0f47,	// range
	0xf0f0ffff, 0x0f49, 0x0f69,	// range
	0xf0f0ffff, 0x0f71, 0x0f76,	// range
	0x0f78,
	0xf0f0ffff, 0x0f7a, 0x0f7d,	// range
	0xf0f0ffff, 0x0f80, 0x0f81,	// range
	0xf0f0ffff, 0x0f90, 0x0f97,	// range
	0xf0f0ffff, 0x0f99, 0x0fb9,	// range
};
#define BO_LANG_ORTH_SZ	22
