
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int fil_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x00c0, 0x00c2,	// range
	0xf0f0ffff, 0x00c8, 0x00ca,	// range
	0xf0f0ffff, 0x00cc, 0x00ce,	// range
	0xf0f0ffff, 0x00d1, 0x00d4,	// range
	0xf0f0ffff, 0x00d9, 0x00db,	// range
	0xf0f0ffff, 0x00e0, 0x00e2,	// range
	0xf0f0ffff, 0x00e8, 0x00ea,	// range
	0xf0f0ffff, 0x00ec, 0x00ee,	// range
	0xf0f0ffff, 0x00f1, 0x00f4,	// range
	0xf0f0ffff, 0x00f9, 0x00fb,	// range
};
#define FIL_LANG_ORTH_SZ	36
