
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int mni_lang_orth_chars[] = {
	0xf0f0ffff, 0x0981, 0x0983,	// range
	0xf0f0ffff, 0x0985, 0x098c,	// range
	0xf0f0ffff, 0x098f, 0x0990,	// range
	0xf0f0ffff, 0x0993, 0x09a8,	// range
	0xf0f0ffff, 0x09aa, 0x09b0,	// range
	0x09b2,
	0xf0f0ffff, 0x09b6, 0x09b9,	// range
	0x09bc,
	0xf0f0ffff, 0x09be, 0x09c4,	// range
	0xf0f0ffff, 0x09c7, 0x09c8,	// range
	0xf0f0ffff, 0x09cb, 0x09cd,	// range
	0xf0f0ffff, 0x09dc, 0x09dd,	// range
	0x09df,
	0xf0f0ffff, 0x0964, 0x09c4,	// range
	0x09bd,
	0x09ce,
	0xf0f0ffff, 0x09e6, 0x09ef,	// range
	0x09f1,
};
#define MNI_LANG_ORTH_SZ	42
