
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ve_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x1e12, 0x1e13,	// range
	0xf0f0ffff, 0x1e3c, 0x1e3d,	// range
	0xf0f0ffff, 0x1e44, 0x1e45,	// range
	0xf0f0ffff, 0x1e4a, 0x1e4b,	// range
	0xf0f0ffff, 0x1e70, 0x1e71,	// range
};
#define VE_LANG_ORTH_SZ	21
