
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int lo_lang_orth_chars[] = {
	0xf0f0ffff, 0x0e81, 0x0e82,	// range
	0x0e84,
	0xf0f0ffff, 0x0e87, 0x0e88,	// range
	0x0e8a,
	0x0e8d,
	0xf0f0ffff, 0x0e94, 0x0e97,	// range
	0xf0f0ffff, 0x0e99, 0x0e9f,	// range
	0xf0f0ffff, 0x0ea1, 0x0ea3,	// range
	0x0ea5,
	0x0ea7,
	0xf0f0ffff, 0x0eaa, 0x0eab,	// range
	0xf0f0ffff, 0x0ead, 0x0eb9,	// range
	0xf0f0ffff, 0x0ebb, 0x0ebd,	// range
	0xf0f0ffff, 0x0ec0, 0x0ec4,	// range
	0x0ec6,
	0xf0f0ffff, 0x0ec8, 0x0ecd,	// range
	0xf0f0ffff, 0x0edc, 0x0edd,	// range
};
#define LO_LANG_ORTH_SZ	39
