
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int doi_lang_orth_chars[] = {
	0xf0f0ffff, 0x0902, 0x0903,	// range
	0xf0f0ffff, 0x0905, 0x090c,	// range
	0xf0f0ffff, 0x090f, 0x0910,	// range
	0xf0f0ffff, 0x0913, 0x0928,	// range
	0xf0f0ffff, 0x092a, 0x0930,	// range
	0x0932,
	0xf0f0ffff, 0x0935, 0x0939,	// range
	0xf0f0ffff, 0x093c, 0x0944,	// range
	0xf0f0ffff, 0x0947, 0x0948,	// range
	0xf0f0ffff, 0x094b, 0x094d,	// range
	0xf0f0ffff, 0x0950, 0x0952,	// range
	0xf0f0ffff, 0x095b, 0x096f,	// range
};
#define DOI_LANG_ORTH_SZ	34
