
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int und_zsye_lang_orth_chars[] = {
	0xf0f0ffff, 0x231a, 0x231b,	// range
	0xf0f0ffff, 0x23e9, 0x23ec,	// range
	0x23f0,
	0x23f3,
	0xf0f0ffff, 0x25fd, 0x25fe,	// range
	0xf0f0ffff, 0x2614, 0x2615,	// range
	0xf0f0ffff, 0x2648, 0x2653,	// range
	0x267f,
	0x2693,
	0x26a1,
	0xf0f0ffff, 0x26aa, 0x26ab,	// range
	0xf0f0ffff, 0x26bd, 0x26be,	// range
	0xf0f0ffff, 0x26c4, 0x26c5,	// range
	0x26ce,
	0x26d4,
	0x26ea,
	0xf0f0ffff, 0x26f2, 0x26f3,	// range
	0x26f5,
	0x26fa,
	0x26fd,
	0x2705,
	0xf0f0ffff, 0x270a, 0x270b,	// range
	0x2728,
	0x274c,
	0x274e,
	0xf0f0ffff, 0x2753, 0x2755,	// range
	0x2757,
	0xf0f0ffff, 0x2795, 0x2797,	// range
	0x27b0,
	0x27bf,
	0xf0f0ffff, 0x2b1b, 0x2b1c,	// range
	0x2b50,
	0x2b55,
	0x1f004,
	0x1f0cf,
	0x1f18e,
	0xf0f0ffff, 0x1f191, 0x1f19a,	// range
	0xf0f0ffff, 0x1f1e6, 0x1f1ff,	// range
	0x1f201,
	0x1f21a,
	0x1f22f,
	0xf0f0ffff, 0x1f232, 0x1f236,	// range
	0xf0f0ffff, 0x1f238, 0x1f23a,	// range
	0xf0f0ffff, 0x1f250, 0x1f251,	// range
	0xf0f0ffff, 0x1f300, 0x1f320,	// range
	0xf0f0ffff, 0x1f330, 0x1f335,	// range
	0xf0f0ffff, 0x1f337, 0x1f37c,	// range
	0xf0f0ffff, 0x1f380, 0x1f393,	// range
	0xf0f0ffff, 0x1f3a0, 0x1f3c4,	// range
	0xf0f0ffff, 0x1f3c6, 0x1f3ca,	// range
	0xf0f0ffff, 0x1f3e0, 0x1f3f0,	// range
	0x1f440,
	0xf0f0ffff, 0x1f442, 0x1f4f7,	// range
	0xf0f0ffff, 0x1f4f9, 0x1f4fc,	// range
	0xf0f0ffff, 0x1f500, 0x1f53d,	// range
	0xf0f0ffff, 0x1f550, 0x1f567,	// range
	0xf0f0ffff, 0x1f5fb, 0x1f5ff,	// range
	0xf0f0ffff, 0x1f601, 0x1f610,	// range
	0xf0f0ffff, 0x1f612, 0x1f614,	// range
	0x1f616,
	0x1f618,
	0x1f61a,
	0xf0f0ffff, 0x1f61c, 0x1f61e,	// range
	0xf0f0ffff, 0x1f620, 0x1f625,	// range
	0xf0f0ffff, 0x1f628, 0x1f62b,	// range
	0x1f62d,
	0xf0f0ffff, 0x1f630, 0x1f633,	// range
	0xf0f0ffff, 0x1f635, 0x1f640,	// range
	0xf0f0ffff, 0x1f645, 0x1f64f,	// range
};
#define UND_ZSYE_LANG_ORTH_SZ	145
