
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int bh_lang_orth_chars[] = {
	0xf0f0ffff, 0x0905, 0x0914,	// range
	0xf0f0ffff, 0x0915, 0x0939,	// range
	0xf0f0ffff, 0x093f, 0x094c,	// range
	0x094d,
};
#define BH_LANG_ORTH_SZ	10
