
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int shs_lang_orth_chars[] = {
	0x0037,
	0x0041,
	0x0043,
	0x0045,
	0xf0f0ffff, 0x0047, 0x0049,	// range
	0xf0f0ffff, 0x004b, 0x0055,	// range
	0xf0f0ffff, 0x0057, 0x0059,	// range
	0x0061,
	0x0063,
	0x0065,
	0xf0f0ffff, 0x0067, 0x0069,	// range
	0xf0f0ffff, 0x006b, 0x0075,	// range
	0xf0f0ffff, 0x0077, 0x0079,	// range
	0x00c1,
	0x00c9,
	0x00cd,
	0x00e1,
	0x00e9,
	0x00ed,
	0x0313,
};
#define SHS_LANG_ORTH_SZ	32
