
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int yo_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c0,
	0x00c1,
	0x00c2,
	0x00c3,
	0x00c8,
	0x00c9,
	0x00ca,
	0x00cc,
	0x00cd,
	0x00ce,
	0x00d2,
	0x00d3,
	0x00d4,
	0x00d5,
	0x00d9,
	0x00da,
	0x00db,
	0x00e0,
	0x00e1,
	0x00e2,
	0x00e3,
	0x00e8,
	0x00e9,
	0x00ea,
	0x00ec,
	0x00ed,
	0x00ee,
	0x00f2,
	0x00f3,
	0x00f4,
	0x00f5,
	0x00f9,
	0x00fa,
	0x00fb,
	0xf0f0ffff, 0x011a, 0x011b,	// range
	0xf0f0ffff, 0x0128, 0x0129,	// range
	0xf0f0ffff, 0x0143, 0x0144,	// range
	0xf0f0ffff, 0x0168, 0x0169,	// range
	0xf0f0ffff, 0x01cd, 0x01ce,	// range
	0xf0f0ffff, 0x01cf, 0x01d0,	// range
	0xf0f0ffff, 0x01d1, 0x01d2,	// range
	0xf0f0ffff, 0x01d3, 0x01d4,	// range
	0xf0f0ffff, 0x01f8, 0x01f9,	// range
	0x0300,
	0x0301,
	0x0302,
	0x0303,
	0x030c,
	0xf0f0ffff, 0x1e3e, 0x1e3f,	// range
	0xf0f0ffff, 0x1e62, 0x1e63,	// range
	0xf0f0ffff, 0x1eb8, 0x1eb9,	// range
	0xf0f0ffff, 0x1ebc, 0x1ebd,	// range
	0xf0f0ffff, 0x1ecc, 0x1ecd,	// range
};
#define YO_LANG_ORTH_SZ	87
