
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int csb_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c3,
	0x00c9,
	0x00cb,
	0xf0f0ffff, 0x00d2, 0x00d4,	// range
	0x00d9,
	0x00e3,
	0x00e9,
	0x00eb,
	0xf0f0ffff, 0x00f2, 0x00f4,	// range
	0x00f9,
	0xf0f0ffff, 0x0104, 0x0105,	// range
	0xf0f0ffff, 0x0141, 0x0144,	// range
	0xf0f0ffff, 0x017b, 0x017c,	// range
};
#define CSB_LANG_ORTH_SZ	29
