
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ug_lang_orth_chars[] = {
	0xf0f0ffff, 0x0626, 0x0628,	// range
	0x062a,
	0x062c,
	0xf0f0ffff, 0x062e, 0x062f,	// range
	0xf0f0ffff, 0x0631, 0x0634,	// range
	0x063a,
	0xf0f0ffff, 0x0641, 0x0646,	// range
	0xf0f0ffff, 0x0648, 0x064a,	// range
	0x067e,
	0x0686,
	0x0698,
	0x06ad,
	0x06af,
	0x06be,
	0xf0f0ffff, 0x06c6, 0x06c8,	// range
	0x06cb,
	0x06d0,
	0x06d5,
};
#define UG_LANG_ORTH_SZ	30
