package org.eink_onyx_reflections;

public enum DeviceType {
	none,
	imx6,		// Freescale i.MX6
	imx7,		// Freescale i.MX7
	rk3026,		// Rockchip RK3026
	rk31xx,		// Rockchip RK31XX
	rk32xx,		// Rockchip RK3288
	rk33xx,		// Rockchip RK3368
	sdm			// Snapdragon 625, 630, 660, ...
}
