
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int kn_lang_orth_chars[] = {
	0xf0f0ffff, 0x0c82, 0x0c83,	// range
	0xf0f0ffff, 0x0c85, 0x0c8c,	// range
	0xf0f0ffff, 0x0c8e, 0x0c90,	// range
	0xf0f0ffff, 0x0c92, 0x0ca8,	// range
	0xf0f0ffff, 0x0caa, 0x0cb3,	// range
	0xf0f0ffff, 0x0cb5, 0x0cb9,	// range
	0xf0f0ffff, 0x0cbe, 0x0cc4,	// range
	0xf0f0ffff, 0x0cc6, 0x0cc8,	// range
	0xf0f0ffff, 0x0cca, 0x0ccd,	// range
	0xf0f0ffff, 0x0cd5, 0x0cd6,	// range
	0x0cde,
	0xf0f0ffff, 0x0ce0, 0x0ce1,	// range
};
#define KN_LANG_ORTH_SZ	34
