
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int bg_lang_orth_chars[] = {
	0xf0f0ffff, 0x0410, 0x042a,	// range
	0x042c,
	0xf0f0ffff, 0x042e, 0x042f,	// range
	0xf0f0ffff, 0x0430, 0x044a,	// range
	0x044c,
	0xf0f0ffff, 0x044e, 0x044f,	// range
};
#define BG_LANG_ORTH_SZ	14
