
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int te_lang_orth_chars[] = {
	0xf0f0ffff, 0x0c01, 0x0c03,	// range
	0xf0f0ffff, 0x0c05, 0x0c0c,	// range
	0xf0f0ffff, 0x0c0e, 0x0c10,	// range
	0xf0f0ffff, 0x0c12, 0x0c28,	// range
	0xf0f0ffff, 0x0c2a, 0x0c33,	// range
	0xf0f0ffff, 0x0c35, 0x0c39,	// range
	0xf0f0ffff, 0x0c3e, 0x0c44,	// range
	0xf0f0ffff, 0x0c46, 0x0c48,	// range
	0xf0f0ffff, 0x0c4a, 0x0c4d,	// range
	0xf0f0ffff, 0x0c55, 0x0c56,	// range
	0xf0f0ffff, 0x0c60, 0x0c61,	// range
};
#define TE_LANG_ORTH_SZ	33
