
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int az_ir_lang_orth_chars[] = {
	0xf0f0ffff, 0x0621, 0x0624,	// range
	0xf0f0ffff, 0x0626, 0x0628,	// range
	0x0629,
	0xf0f0ffff, 0x062a, 0x063a,	// range
	0xf0f0ffff, 0x0641, 0x0642,	// range
	0xf0f0ffff, 0x0644, 0x0648,	// range
	0x064b,
	0x0654,
	0x067e,
	0x0686,
	0x0698,
	0x06a9,
	0x06af,
	0x06cc,
	0x0654,
};
#define AZ_IR_LANG_ORTH_SZ	25
