
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int sr_lang_orth_chars[] = {
	0x0402,
	0xf0f0ffff, 0x0408, 0x040b,	// range
	0x040f,
	0xf0f0ffff, 0x0410, 0x0418,	// range
	0xf0f0ffff, 0x041a, 0x0428,	// range
	0xf0f0ffff, 0x0430, 0x0438,	// range
	0xf0f0ffff, 0x043a, 0x0448,	// range
	0x0452,
	0xf0f0ffff, 0x0458, 0x045b,	// range
	0x045f,
};
#define SR_LANG_ORTH_SZ	22
