
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int or_lang_orth_chars[] = {
	0xf0f0ffff, 0x0b01, 0x0b03,	// range
	0xf0f0ffff, 0x0b05, 0x0b0c,	// range
	0xf0f0ffff, 0x0b0f, 0x0b10,	// range
	0xf0f0ffff, 0x0b13, 0x0b28,	// range
	0xf0f0ffff, 0x0b2a, 0x0b30,	// range
	0xf0f0ffff, 0x0b32, 0x0b33,	// range
	0xf0f0ffff, 0x0b36, 0x0b39,	// range
	0xf0f0ffff, 0x0b3c, 0x0b43,	// range
	0xf0f0ffff, 0x0b47, 0x0b48,	// range
	0xf0f0ffff, 0x0b4b, 0x0b4d,	// range
	0xf0f0ffff, 0x0b56, 0x0b57,	// range
	0xf0f0ffff, 0x0b5c, 0x0b5d,	// range
	0xf0f0ffff, 0x0b5f, 0x0b61,	// range
};
#define OR_LANG_ORTH_SZ	39
