
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int si_lang_orth_chars[] = {
	0xf0f0ffff, 0x0d82, 0x0d83,	// range
	0xf0f0ffff, 0x0d85, 0x0d8d,	// range
	0xf0f0ffff, 0x0d91, 0x0d96,	// range
	0xf0f0ffff, 0x0d9a, 0x0da5,	// range
	0xf0f0ffff, 0x0da7, 0x0db1,	// range
	0xf0f0ffff, 0x0db3, 0x0dbb,	// range
	0x0dbd,
	0xf0f0ffff, 0x0dc0, 0x0dc6,	// range
	0x0dca,
	0xf0f0ffff, 0x0dcf, 0x0dd4,	// range
	0x0dd6,
	0xf0f0ffff, 0x0dd8, 0x0dde,	// range
	0x0df2,
};
#define SI_LANG_ORTH_SZ	31
