
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ta_lang_orth_chars[] = {
	0x0b83,
	0xf0f0ffff, 0x0b85, 0x0b8a,	// range
	0xf0f0ffff, 0x0b8e, 0x0b90,	// range
	0xf0f0ffff, 0x0b92, 0x0b95,	// range
	0xf0f0ffff, 0x0b99, 0x0b9a,	// range
	0x0b9c,
	0xf0f0ffff, 0x0b9e, 0x0b9f,	// range
	0xf0f0ffff, 0x0ba3, 0x0ba4,	// range
	0xf0f0ffff, 0x0ba8, 0x0baa,	// range
	0xf0f0ffff, 0x0bae, 0x0bb5,	// range
	0xf0f0ffff, 0x0bb7, 0x0bb9,	// range
	0xf0f0ffff, 0x0bbe, 0x0bc2,	// range
	0xf0f0ffff, 0x0bc6, 0x0bc8,	// range
	0xf0f0ffff, 0x0bca, 0x0bcd,	// range
	0x0bd7,
};
#define TA_LANG_ORTH_SZ	39
