
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ee_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x00c0, 0x00c1,	// range
	0xf0f0ffff, 0x00c8, 0x00c9,	// range
	0xf0f0ffff, 0x00cc, 0x00cd,	// range
	0xf0f0ffff, 0x00d2, 0x00d3,	// range
	0xf0f0ffff, 0x00d9, 0x00da,	// range
	0xf0f0ffff, 0x00e0, 0x00e1,	// range
	0xf0f0ffff, 0x00e8, 0x00e9,	// range
	0xf0f0ffff, 0x00ec, 0x00ed,	// range
	0xf0f0ffff, 0x00f2, 0x00f3,	// range
	0xf0f0ffff, 0x00f9, 0x00fa,	// range
	0xf0f0ffff, 0x011a, 0x011b,	// range
	0xf0f0ffff, 0x014a, 0x014b,	// range
	0x0186,
	0x0189,
	0xf0f0ffff, 0x0190, 0x0192,	// range
	0x0194,
	0x01b2,
	0xf0f0ffff, 0x01cd, 0x01d4,	// range
	0x0254,
	0x0256,
	0x025b,
	0x0263,
	0x028b,
	0xf0f0ffff, 0x0300, 0x0301,	// range
	0x030c,
};
#define EE_LANG_ORTH_SZ	61
