
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ab_lang_orth_chars[] = {
	0x0401,
	0x040f,
	0xf0f0ffff, 0x0410, 0x044f,	// range
	0x0451,
	0x045f,
	0xf0f0ffff, 0x049e, 0x049f,	// range
	0xf0f0ffff, 0x04a6, 0x04a9,	// range
	0xf0f0ffff, 0x04ac, 0x04ad,	// range
	0xf0f0ffff, 0x04b2, 0x04b7,	// range
	0xf0f0ffff, 0x04bc, 0x04bf,	// range
	0x04d8,
	0x04d9,
	0xf0f0ffff, 0x04e0, 0x04e1,	// range
};
#define AB_LANG_ORTH_SZ	27
