// ISO-639-3 Language registry.
// License: Public Domain
// This file is autogenerated.
// Based on data from https://iso639-3.sil.org/code_tables/download_tables
// file: iso-639-3.tab
// url: https://iso639-3.sil.org/sites/iso639-3/files/downloads/iso-639-3.tab

#ifndef ISO639_3_DATA_H
#define ISO639_3_DATA_H

#ifdef __cplusplus
extern "C" {
#endif

#define ISO639_3_DATA_SZ	7894
#define ISO639_3_UND_INDEX	6780

struct iso639_3_rec {
	/**
	 * The three-letter 639-3 identifier.
	 */
	const char* id;
	/**
	 * Equivalent 639-2 identifier of the bibliographic application code set, if there is one.
	 */
	const char* part2b;
	/**
	 * Equivalent 639-2 identifier of the terminology application code set, if there is one.
	 */
	const char* part2t;
	/**
	 * Equivalent 639-1 identifier, if there is one.
	 */
	const char* part1;
	/**
	 * Reference language name.
	 */
	const char* ref_name;
};

extern const struct iso639_3_rec iso639_3_data[];

#ifdef __cplusplus
}
#endif

#endif	// ISO639_3_DATA_H
