
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int kab_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x010c, 0x010d,	// range
	0x0190,
	0x0194,
	0xf0f0ffff, 0x01e6, 0x01e7,	// range
	0x025b,
	0x0263,
	0xf0f0ffff, 0x1e0c, 0x1e0d,	// range
	0xf0f0ffff, 0x1e24, 0x1e25,	// range
	0xf0f0ffff, 0x1e62, 0x1e63,	// range
	0xf0f0ffff, 0x1e6c, 0x1e6d,	// range
	0xf0f0ffff, 0x1e92, 0x1e93,	// range
};
#define KAB_LANG_ORTH_SZ	31
