
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int cy_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c2,
	0x00c9,
	0x00ca,
	0x00ce,
	0x00cf,
	0x00d4,
	0x00e2,
	0x00e9,
	0x00ea,
	0x00ee,
	0x00ef,
	0x00f4,
	0x00ff,
	0x0174,
	0x0175,
	0x0176,
	0x0177,
	0x0178,
	0x1e80,
	0x1e81,
	0x1e82,
	0x1e83,
	0x1e84,
	0x1e85,
	0x1ef2,
	0x1ef3,
};
#define CY_LANG_ORTH_SZ	32
