
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int kk_lang_orth_chars[] = {
	0xf0f0ffff, 0x0410, 0x044f,	// range
	0x0456,
	0xf0f0ffff, 0x0492, 0x0493,	// range
	0xf0f0ffff, 0x049a, 0x049b,	// range
	0xf0f0ffff, 0x04a2, 0x04a3,	// range
	0xf0f0ffff, 0x04ba, 0x04bb,	// range
	0xf0f0ffff, 0x04d8, 0x04d9,	// range
	0xf0f0ffff, 0x04e8, 0x04e9,	// range
};
#define KK_LANG_ORTH_SZ	22
