
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int km_lang_orth_chars[] = {
	0xf0f0ffff, 0x1780, 0x179c,	// range
	0xf0f0ffff, 0x179f, 0x17a2,	// range
	0xf0f0ffff, 0x17a5, 0x17a7,	// range
	0xf0f0ffff, 0x17a9, 0x17b3,	// range
	0xf0f0ffff, 0x17b6, 0x17c5,	// range
};
#define KM_LANG_ORTH_SZ	15
