
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int tig_lang_orth_chars[] = {
	0xf0f0ffff, 0x1200, 0x1206,	// range
	0xf0f0ffff, 0x1208, 0x1216,	// range
	0xf0f0ffff, 0x1218, 0x121f,	// range
	0xf0f0ffff, 0x1228, 0x1230,	// range
	0xf0f0ffff, 0x1238, 0x1246,	// range
	0x1248,
	0xf0f0ffff, 0x124a, 0x124d,	// range
	0xf0f0ffff, 0x1260, 0x126e,	// range
	0xf0f0ffff, 0x1270, 0x127f,	// range
	0xf0f0ffff, 0x1290, 0x1297,	// range
	0xf0f0ffff, 0x12a0, 0x12a6,	// range
	0xf0f0ffff, 0x12a8, 0x12ae,	// range
	0x12b0,
	0xf0f0ffff, 0x12b2, 0x12b5,	// range
	0xf0f0ffff, 0x12c8, 0x12ce,	// range
	0xf0f0ffff, 0x12d0, 0x12d6,	// range
	0xf0f0ffff, 0x12d8, 0x12df,	// range
	0xf0f0ffff, 0x12e8, 0x12ee,	// range
	0xf0f0ffff, 0x12f0, 0x12f7,	// range
	0xf0f0ffff, 0x1300, 0x130e,	// range
	0x1310,
	0xf0f0ffff, 0x1312, 0x1315,	// range
	0xf0f0ffff, 0x1320, 0x133f,	// range
	0xf0f0ffff, 0x1348, 0x1356,	// range
};
#define TIG_LANG_ORTH_SZ	66
