
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int my_lang_orth_chars[] = {
	0xf0f0ffff, 0x1000, 0x1020,	// range
	0x1021,
	0xf0f0ffff, 0x1023, 0x1027,	// range
	0xf0f0ffff, 0x1029, 0x102a,	// range
	0xf0f0ffff, 0x102c, 0x1032,	// range
};
#define MY_LANG_ORTH_SZ	13
