
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int kl_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c1,
	0x00c2,
	0x00c3,
	0x00c5,
	0x00c6,
	0x00ca,
	0x00cd,
	0x00ce,
	0x00d4,
	0x00d8,
	0x00da,
	0x00db,
	0x00e1,
	0x00e2,
	0x00e3,
	0x00e5,
	0x00e6,
	0x00ea,
	0x00ed,
	0x00ee,
	0x00f4,
	0x00f8,
	0x00fa,
	0x00fb,
	0xf0f0ffff, 0x0128, 0x0129,	// range
	0x0138,
	0xf0f0ffff, 0x0168, 0x0169,	// range
};
#define KL_LANG_ORTH_SZ	37
