
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int mn_cn_lang_orth_chars[] = {
	0xf0f0ffff, 0x1820, 0x1842,	// range
	0xf0f0ffff, 0x1843, 0x185c,	// range
	0xf0f0ffff, 0x185d, 0x1872,	// range
	0xf0f0ffff, 0x1873, 0x1877,	// range
	0xf0f0ffff, 0x1880, 0x18a9,	// range
};
#define MN_CN_LANG_ORTH_SZ	15
