
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int nb_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c0,
	0x00c5,
	0x00c6,
	0x00c9,
	0x00ca,
	0x00d2,
	0x00d3,
	0x00d4,
	0x00d8,
	0x00e0,
	0x00e5,
	0x00e6,
	0x00e9,
	0x00ea,
	0x00f2,
	0x00f3,
	0x00f4,
	0x00f8,
};
#define NB_LANG_ORTH_SZ	24
