
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int bin_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c0,
	0x00c1,
	0x00c8,
	0x00c9,
	0x00cc,
	0x00cd,
	0x00d2,
	0x00d3,
	0x00d9,
	0x00da,
	0x00e0,
	0x00e1,
	0x00e8,
	0x00e9,
	0x00ec,
	0x00ed,
	0x00f2,
	0x00f3,
	0x00f9,
	0x00fa,
	0x1eb8,
	0x1eb9,
	0x1ecc,
	0x1ecd,
	0x0300,
	0x0301,
};
#define BIN_LANG_ORTH_SZ	32
