
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ln_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x00c1, 0x00c2,	// range
	0xf0f0ffff, 0x00c9, 0x00ca,	// range
	0xf0f0ffff, 0x00cd, 0x00ce,	// range
	0xf0f0ffff, 0x00d3, 0x00d4,	// range
	0xf0f0ffff, 0x00da, 0x00db,	// range
	0xf0f0ffff, 0x00e1, 0x00e2,	// range
	0xf0f0ffff, 0x00e9, 0x00ea,	// range
	0xf0f0ffff, 0x00ed, 0x00ee,	// range
	0xf0f0ffff, 0x00f3, 0x00f4,	// range
	0xf0f0ffff, 0x00fa, 0x00fb,	// range
	0xf0f0ffff, 0x011a, 0x011b,	// range
	0x0186,
	0x0190,
	0x0254,
	0x025b,
	0xf0f0ffff, 0x0301, 0x0302,	// range
	0x030c,
};
#define LN_LANG_ORTH_SZ	47
