
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int vi_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x00c0, 0x00c3,	// range
	0xf0f0ffff, 0x00c8, 0x00ca,	// range
	0xf0f0ffff, 0x00cc, 0x00cd,	// range
	0xf0f0ffff, 0x00d2, 0x00d5,	// range
	0xf0f0ffff, 0x00d9, 0x00da,	// range
	0x00dd,
	0xf0f0ffff, 0x00e0, 0x00e3,	// range
	0xf0f0ffff, 0x00e8, 0x00ea,	// range
	0xf0f0ffff, 0x00ec, 0x00ed,	// range
	0xf0f0ffff, 0x00f2, 0x00f5,	// range
	0xf0f0ffff, 0x00f9, 0x00fa,	// range
	0x00fd,
	0xf0f0ffff, 0x0102, 0x0103,	// range
	0xf0f0ffff, 0x0110, 0x0111,	// range
	0xf0f0ffff, 0x0128, 0x0129,	// range
	0xf0f0ffff, 0x0168, 0x0169,	// range
	0xf0f0ffff, 0x01a0, 0x01a1,	// range
	0xf0f0ffff, 0x01af, 0x01b0,	// range
	0xf0f0ffff, 0x0300, 0x0303,	// range
	0x0306,
	0x0309,
	0x031b,
	0x0323,
	0xf0f0ffff, 0x1ea0, 0x1ef9,	// range
};
#define VI_LANG_ORTH_SZ	66
