
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int sh_lang_orth_chars[] = {
	0xf0f0ffff, 0x0410, 0x044f,	// range
	0x0492,
	0x0493,
	0x0498,
	0x0499,
	0x04a0,
	0x04a1,
	0x04a2,
	0x04a3,
	0x04aa,
	0x04ab,
	0x04ae,
	0x04af,
	0x04ba,
	0x04bb,
	0x04d8,
	0x04d9,
	0x04e8,
	0x04e9,
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0xf0f0ffff, 0x0106, 0x0107,	// range
	0xf0f0ffff, 0x010c, 0x010d,	// range
	0xf0f0ffff, 0x0110, 0x0111,	// range
	0xf0f0ffff, 0x0160, 0x0161,	// range
	0xf0f0ffff, 0x017d, 0x017e,	// range
	0x0402,
	0xf0f0ffff, 0x0408, 0x040b,	// range
	0x040f,
	0xf0f0ffff, 0x0410, 0x0418,	// range
	0xf0f0ffff, 0x041a, 0x0428,	// range
	0xf0f0ffff, 0x0430, 0x0438,	// range
	0xf0f0ffff, 0x043a, 0x0448,	// range
	0x0452,
	0xf0f0ffff, 0x0458, 0x045b,	// range
	0x045f,
};
#define SH_LANG_ORTH_SZ	64
