
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int bs_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x0106,
	0x0107,
	0x010c,
	0x010d,
	0x0110,
	0x0111,
	0x0160,
	0x0161,
	0x017d,
	0x017e,
};
#define BS_LANG_ORTH_SZ	16
