
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int mk_lang_orth_chars[] = {
	0x0400,
	0x0403,
	0x0405,
	0xf0f0ffff, 0x0408, 0x040a,	// range
	0x040c,
	0x040d,
	0x040f,
	0xf0f0ffff, 0x0410, 0x0418,	// range
	0xf0f0ffff, 0x041a, 0x0428,	// range
	0x0450,
	0x0453,
	0x0455,
	0xf0f0ffff, 0x0458, 0x045a,	// range
	0x045c,
	0x045d,
	0x045f,
};
#define MK_LANG_ORTH_SZ	24
