
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ber_ma_lang_orth_chars[] = {
	0xf0f0ffff, 0x2d30, 0x2d31,	// range
	0x2d33,
	0x2d37,
	0x2d39,
	0xf0f0ffff, 0x2d3b, 0x2d3d,	// range
	0x2d40,
	0xf0f0ffff, 0x2d43, 0x2d45,	// range
	0x2d47,
	0xf0f0ffff, 0x2d49, 0x2d4a,	// range
	0xf0f0ffff, 0x2d4d, 0x2d4f,	// range
	0xf0f0ffff, 0x2d53, 0x2d56,	// range
	0xf0f0ffff, 0x2d59, 0x2d5c,	// range
	0x2d5f,
	0xf0f0ffff, 0x2d61, 0x2d63,	// range
	0x2d65,
	0x2d6f,
};
#define BER_MA_LANG_ORTH_SZ	32
