
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int pa_lang_orth_chars[] = {
	0xf0f0ffff, 0x0a05, 0x0a0a,	// range
	0xf0f0ffff, 0x0a0f, 0x0a10,	// range
	0xf0f0ffff, 0x0a13, 0x0a14,	// range
	0xf0f0ffff, 0x0a15, 0x0a28,	// range
	0xf0f0ffff, 0x0a2a, 0x0a30,	// range
	0xf0f0ffff, 0x0a32, 0x0a33,	// range
	0xf0f0ffff, 0x0a35, 0x0a36,	// range
	0xf0f0ffff, 0x0a38, 0x0a39,	// range
	0x0a3c,
	0xf0f0ffff, 0x0a3e, 0x0a42,	// range
	0xf0f0ffff, 0x0a47, 0x0a48,	// range
	0xf0f0ffff, 0x0a4b, 0x0a4c,	// range
	0x0a4d,
	0xf0f0ffff, 0x0a59, 0x0a5c,	// range
	0xf0f0ffff, 0x0a70, 0x0a74,	// range
};
#define PA_LANG_ORTH_SZ	41
