
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int sd_lang_orth_chars[] = {
	0xf0f0ffff, 0x0621, 0x0622,	// range
	0x0624,
	0xf0f0ffff, 0x0626, 0x0628,	// range
	0xf0f0ffff, 0x062a, 0x063a,	// range
	0xf0f0ffff, 0x0641, 0x0642,	// range
	0xf0f0ffff, 0x0644, 0x0648,	// range
	0x064a,
	0xf0f0ffff, 0x067a, 0x067b,	// range
	0xf0f0ffff, 0x067d, 0x0680,	// range
	0xf0f0ffff, 0x0683, 0x0684,	// range
	0xf0f0ffff, 0x0686, 0x0687,	// range
	0x068a,
	0xf0f0ffff, 0x068c, 0x068d,	// range
	0x068f,
	0x0699,
	0x06a6,
	0xf0f0ffff, 0x06a9, 0x06aa,	// range
	0x06af,
	0x06b1,
	0x06b3,
	0x06bb,
	0x06be,
};
#define SD_LANG_ORTH_SZ	44
