
# Metadata for deploy script

PN="utf8proc"
PV="2.6.1"
# package revision: when patchset is changed (but not version), increase it
# when version changed, reset to "1".
REV="1"
SRCFILE="v${PV}.tar.gz"
SHA512="9a09fe58a63c8c7983b8968b9b2c82f3d188724ed125097ea0de6a18d87569c678e226dc5de183108e47b3f068117e6bb352da1c3d24ebfd8018fb0b2346592e"

URL="https://github.com/JuliaStrings/utf8proc/archive/refs/tags/v${PV}.tar.gz"

SOURCESDIR="${PN}-${PV}"
