
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ru_lang_orth_chars[] = {
	0x0401,
	0xf0f0ffff, 0x0410, 0x044f,	// range
	0x0451,
};
#define RU_LANG_ORTH_SZ	5
