
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int cs_lang_orth_chars[] = {
	0xf0f0ffff, 0x0041, 0x005a,	// range
	0xf0f0ffff, 0x0061, 0x007a,	// range
	0x00c1,
	0x00c9,
	0x00cd,
	0x00d3,
	0x00da,
	0x00dd,
	0x00e1,
	0x00e9,
	0x00ed,
	0x00f3,
	0x00fa,
	0x00fd,
	0x010c,
	0x010d,
	0x010e,
	0x010f,
	0x011a,
	0x011b,
	0x0147,
	0x0148,
	0x0158,
	0x0159,
	0x0160,
	0x0161,
	0x0164,
	0x0165,
	0x016e,
	0x016f,
	0x017d,
	0x017e,
};
#define CS_LANG_ORTH_SZ	36
