// ISO 15924 Codes for the representation of names of scripts.
// License: Public Domain.
// This file is autogenerated.
// Based on data from: https://unicode.org/iso15924/
// file: iso15924-codes.html
// url: https://unicode.org/iso15924/iso15924-codes.html

#include "iso-15924_data.h"

#include <stddef.h>

const struct iso15924_rec iso15924_data[] = {
	{"Adlm", "Adlam", "Adlam", 166},
	{"Afak", "Afaka", NULL, 439},
	{"Aghb", "Caucasian Albanian", "Caucasian_Albanian", 239},
	{"Ahom", "Ahom, Tai Ahom", "Ahom", 338},
	{"Arab", "Arabic", "Arabic", 160},
	{"Aran", "Arabic (Nastaliq variant)", NULL, 161},
	{"Armi", "Imperial Aramaic", "Imperial_Aramaic", 124},
	{"Armn", "Armenian", "Armenian", 230},
	{"Avst", "Avestan", "Avestan", 134},
	{"Bali", "Balinese", "Balinese", 360},
	{"Bamu", "Bamum", "Bamum", 435},
	{"Bass", "Bassa Vah", "Bassa_Vah", 259},
	{"Batk", "Batak", "Batak", 365},
	{"Beng", "Bengali (Bangla)", "Bengali", 325},
	{"Bhks", "Bhaiksuki", "Bhaiksuki", 334},
	{"Blis", "Blissymbols", NULL, 550},
	{"Bopo", "Bopomofo", "Bopomofo", 285},
	{"Brah", "Brahmi", "Brahmi", 300},
	{"Brai", "Braille", "Braille", 570},
	{"Bugi", "Buginese", "Buginese", 367},
	{"Buhd", "Buhid", "Buhid", 372},
	{"Cakm", "Chakma", "Chakma", 349},
	{"Cans", "Unified Canadian Aboriginal Syllabics", "Canadian_Aboriginal", 440},
	{"Cari", "Carian", "Carian", 201},
	{"Cham", "Cham", "Cham", 358},
	{"Cher", "Cherokee", "Cherokee", 445},
	{"Chrs", "Chorasmian", "Chorasmian", 109},
	{"Cirt", "Cirth", NULL, 291},
	{"Copt", "Coptic", "Coptic", 204},
	{"Cpmn", "Cypro-Minoan", NULL, 402},
	{"Cprt", "Cypriot syllabary", "Cypriot", 403},
	{"Cyrl", "Cyrillic", "Cyrillic", 220},
	{"Cyrs", "Cyrillic (Old Church Slavonic variant)", NULL, 221},
	{"Deva", "Devanagari (Nagari)", "Devanagari", 315},
	{"Diak", "Dives Akuru", "Dives_Akuru", 342},
	{"Dogr", "Dogra", "Dogra", 328},
	{"Dsrt", "Deseret (Mormon)", "Deseret", 250},
	{"Dupl", "Duployan shorthand, Duployan stenography", "Duployan", 755},
	{"Egyd", "Egyptian demotic", NULL, 70},
	{"Egyh", "Egyptian hieratic", NULL, 60},
	{"Egyp", "Egyptian hieroglyphs", "Egyptian_Hieroglyphs", 50},
	{"Elba", "Elbasan", "Elbasan", 226},
	{"Elym", "Elymaic", "Elymaic", 128},
	{"Ethi", "Ethiopic (Geʻez)", "Ethiopic", 430},
	{"Geok", "Khutsuri (Asomtavruli and Nuskhuri)", "Georgian", 241},
	{"Geor", "Georgian (Mkhedruli and Mtavruli)", "Georgian", 240},
	{"Glag", "Glagolitic", "Glagolitic", 225},
	{"Gong", "Gunjala Gondi", "Gunjala_Gondi", 312},
	{"Gonm", "Masaram Gondi", "Masaram_Gondi", 313},
	{"Goth", "Gothic", "Gothic", 206},
	{"Gran", "Grantha", "Grantha", 343},
	{"Grek", "Greek", "Greek", 200},
	{"Gujr", "Gujarati", "Gujarati", 320},
	{"Guru", "Gurmukhi", "Gurmukhi", 310},
	{"Hanb", "Han with Bopomofo (alias for Han + Bopomofo)", NULL, 503},
	{"Hang", "Hangul (Hangŭl, Hangeul)", "Hangul", 286},
	{"Hani", "Han (Hanzi, Kanji, Hanja)", "Han", 500},
	{"Hano", "Hanunoo (Hanunóo)", "Hanunoo", 371},
	{"Hans", "Han (Simplified variant)", NULL, 501},
	{"Hant", "Han (Traditional variant)", NULL, 502},
	{"Hatr", "Hatran", "Hatran", 127},
	{"Hebr", "Hebrew", "Hebrew", 125},
	{"Hira", "Hiragana", "Hiragana", 410},
	{"Hluw", "Anatolian Hieroglyphs (Luwian Hieroglyphs, Hittite Hieroglyphs)", "Anatolian_Hieroglyphs", 80},
	{"Hmng", "Pahawh Hmong", "Pahawh_Hmong", 450},
	{"Hmnp", "Nyiakeng Puachue Hmong", "Nyiakeng_Puachue_Hmong", 451},
	{"Hrkt", "Japanese syllabaries (alias for Hiragana + Katakana)", "Katakana_Or_Hiragana", 412},
	{"Hung", "Old Hungarian (Hungarian Runic)", "Old_Hungarian", 176},
	{"Inds", "Indus (Harappan)", NULL, 610},
	{"Ital", "Old Italic (Etruscan, Oscan, etc.)", "Old_Italic", 210},
	{"Jamo", "Jamo (alias for Jamo subset of Hangul)", NULL, 284},
	{"Java", "Javanese", "Javanese", 361},
	{"Jpan", "Japanese (alias for Han + Hiragana + Katakana)", NULL, 413},
	{"Jurc", "Jurchen", NULL, 510},
	{"Kali", "Kayah Li", "Kayah_Li", 357},
	{"Kana", "Katakana", "Katakana", 411},
	{"Khar", "Kharoshthi", "Kharoshthi", 305},
	{"Khmr", "Khmer", "Khmer", 355},
	{"Khoj", "Khojki", "Khojki", 322},
	{"Kitl", "Khitan large script", NULL, 505},
	{"Kits", "Khitan small script", "Khitan_Small_Script", 288},
	{"Knda", "Kannada", "Kannada", 345},
	{"Kore", "Korean (alias for Hangul + Han)", NULL, 287},
	{"Kpel", "Kpelle", NULL, 436},
	{"Kthi", "Kaithi", "Kaithi", 317},
	{"Lana", "Tai Tham (Lanna)", "Tai_Tham", 351},
	{"Laoo", "Lao", "Lao", 356},
	{"Latf", "Latin (Fraktur variant)", NULL, 217},
	{"Latg", "Latin (Gaelic variant)", NULL, 216},
	{"Latn", "Latin", "Latin", 215},
	{"Leke", "Leke", NULL, 364},
	{"Lepc", "Lepcha (Róng)", "Lepcha", 335},
	{"Limb", "Limbu", "Limbu", 336},
	{"Lina", "Linear A", "Linear_A", 400},
	{"Linb", "Linear B", "Linear_B", 401},
	{"Lisu", "Lisu (Fraser)", "Lisu", 399},
	{"Loma", "Loma", NULL, 437},
	{"Lyci", "Lycian", "Lycian", 202},
	{"Lydi", "Lydian", "Lydian", 116},
	{"Mahj", "Mahajani", "Mahajani", 314},
	{"Maka", "Makasar", "Makasar", 366},
	{"Mand", "Mandaic, Mandaean", "Mandaic", 140},
	{"Mani", "Manichaean", "Manichaean", 139},
	{"Marc", "Marchen", "Marchen", 332},
	{"Maya", "Mayan hieroglyphs", NULL, 90},
	{"Medf", "Medefaidrin (Oberi Okaime, Oberi Ɔkaimɛ)", "Medefaidrin", 265},
	{"Mend", "Mende Kikakui", "Mende_Kikakui", 438},
	{"Merc", "Meroitic Cursive", "Meroitic_Cursive", 101},
	{"Mero", "Meroitic Hieroglyphs", "Meroitic_Hieroglyphs", 100},
	{"Mlym", "Malayalam", "Malayalam", 347},
	{"Modi", "Modi, Moḍī", "Modi", 324},
	{"Mong", "Mongolian", "Mongolian", 145},
	{"Moon", "Moon (Moon code, Moon script, Moon type)", NULL, 218},
	{"Mroo", "Mro, Mru", "Mro", 264},
	{"Mtei", "Meitei Mayek (Meithei, Meetei)", "Meetei_Mayek", 337},
	{"Mult", "Multani", "Multani", 323},
	{"Mymr", "Myanmar (Burmese)", "Myanmar", 350},
	{"Nand", "Nandinagari", "Nandinagari", 311},
	{"Narb", "Old North Arabian (Ancient North Arabian)", "Old_North_Arabian", 106},
	{"Nbat", "Nabataean", "Nabataean", 159},
	{"Newa", "Newa, Newar, Newari, Nepāla lipi", "Newa", 333},
	{"Nkdb", "Naxi Dongba (na²¹ɕi³³ to³³ba²¹, Nakhi Tomba)", NULL, 85},
	{"Nkgb", "Naxi Geba (na²¹ɕi³³ gʌ²¹ba²¹, 'Na-'Khi ²Ggŏ-¹baw, Nakhi Geba)", NULL, 420},
	{"Nkoo", "N’Ko", "Nko", 165},
	{"Nshu", "Nüshu", "Nushu", 499},
	{"Ogam", "Ogham", "Ogham", 212},
	{"Olck", "Ol Chiki (Ol Cemet’, Ol, Santali)", "Ol_Chiki", 261},
	{"Orkh", "Old Turkic, Orkhon Runic", "Old_Turkic", 175},
	{"Orya", "Oriya (Odia)", "Oriya", 327},
	{"Osge", "Osage", "Osage", 219},
	{"Osma", "Osmanya", "Osmanya", 260},
	{"Ougr", "Old Uyghur", NULL, 143},
	{"Palm", "Palmyrene", "Palmyrene", 126},
	{"Pauc", "Pau Cin Hau", "Pau_Cin_Hau", 263},
	{"Pcun", "Proto-Cuneiform", NULL, 15},
	{"Pelm", "Proto-Elamite", NULL, 16},
	{"Perm", "Old Permic", "Old_Permic", 227},
	{"Phag", "Phags-pa", "Phags_Pa", 331},
	{"Phli", "Inscriptional Pahlavi", "Inscriptional_Pahlavi", 131},
	{"Phlp", "Psalter Pahlavi", "Psalter_Pahlavi", 132},
	{"Phlv", "Book Pahlavi", NULL, 133},
	{"Phnx", "Phoenician", "Phoenician", 115},
	{"Plrd", "Miao (Pollard)", "Miao", 282},
	{"Piqd", "Klingon (KLI pIqaD)", NULL, 293},
	{"Prti", "Inscriptional Parthian", "Inscriptional_Parthian", 130},
	{"Psin", "Proto-Sinaitic", NULL, 103},
	{"Qaaa", "Reserved for private use (start)", NULL, 900},
	{"Qabx", "Reserved for private use (end)", NULL, 949},
	{"Ranj", "Ranjana", NULL, 303},
	{"Rjng", "Rejang (Redjang, Kaganga)", "Rejang", 363},
	{"Rohg", "Hanifi Rohingya", "Hanifi_Rohingya", 167},
	{"Roro", "Rongorongo", NULL, 620},
	{"Runr", "Runic", "Runic", 211},
	{"Samr", "Samaritan", "Samaritan", 123},
	{"Sara", "Sarati", NULL, 292},
	{"Sarb", "Old South Arabian", "Old_South_Arabian", 105},
	{"Saur", "Saurashtra", "Saurashtra", 344},
	{"Sgnw", "SignWriting", "SignWriting", 95},
	{"Shaw", "Shavian (Shaw)", "Shavian", 281},
	{"Shrd", "Sharada, Śāradā", "Sharada", 319},
	{"Shui", "Shuishu", NULL, 530},
	{"Sidd", "Siddham, Siddhaṃ, Siddhamātṛkā", "Siddham", 302},
	{"Sind", "Khudawadi, Sindhi", "Khudawadi", 318},
	{"Sinh", "Sinhala", "Sinhala", 348},
	{"Sogd", "Sogdian", "Sogdian", 141},
	{"Sogo", "Old Sogdian", "Old_Sogdian", 142},
	{"Sora", "Sora Sompeng", "Sora_Sompeng", 398},
	{"Soyo", "Soyombo", "Soyombo", 329},
	{"Sund", "Sundanese", "Sundanese", 362},
	{"Sylo", "Syloti Nagri", "Syloti_Nagri", 316},
	{"Syrc", "Syriac", "Syriac", 135},
	{"Syre", "Syriac (Estrangelo variant)", NULL, 138},
	{"Syrj", "Syriac (Western variant)", NULL, 137},
	{"Syrn", "Syriac (Eastern variant)", NULL, 136},
	{"Tagb", "Tagbanwa", "Tagbanwa", 373},
	{"Takr", "Takri, Ṭākrī, Ṭāṅkrī", "Takri", 321},
	{"Tale", "Tai Le", "Tai_Le", 353},
	{"Talu", "New Tai Lue", "New_Tai_Lue", 354},
	{"Taml", "Tamil", "Tamil", 346},
	{"Tang", "Tangut", "Tangut", 520},
	{"Tavt", "Tai Viet", "Tai_Viet", 359},
	{"Telu", "Telugu", "Telugu", 340},
	{"Teng", "Tengwar", NULL, 290},
	{"Tfng", "Tifinagh (Berber)", "Tifinagh", 120},
	{"Tglg", "Tagalog (Baybayin, Alibata)", "Tagalog", 370},
	{"Thaa", "Thaana", "Thaana", 170},
	{"Thai", "Thai", "Thai", 352},
	{"Tibt", "Tibetan", "Tibetan", 330},
	{"Tirh", "Tirhuta", "Tirhuta", 326},
	{"Tnsa", "Tangsa", NULL, 275},
	{"Toto", "Toto", NULL, 294},
	{"Ugar", "Ugaritic", "Ugaritic", 40},
	{"Vaii", "Vai", "Vai", 470},
	{"Visp", "Visible Speech", NULL, 280},
	{"Vith", "Vithkuqi", NULL, 228},
	{"Wara", "Warang Citi (Varang Kshiti)", "Warang_Citi", 262},
	{"Wcho", "Wancho", "Wancho", 283},
	{"Wole", "Woleai", NULL, 480},
	{"Xpeo", "Old Persian", "Old_Persian", 30},
	{"Xsux", "Cuneiform, Sumero-Akkadian", "Cuneiform", 20},
	{"Yezi", "Yezidi", "Yezidi", 192},
	{"Yiii", "Yi", "Yi", 460},
	{"Zanb", "Zanabazar Square (Zanabazarin Dörböljin Useg, Xewtee Dörböljin Bicig, Horizontal Square Script)", "Zanabazar_Square", 339},
	{"Zinh", "Code for inherited script", "Inherited", 994},
	{"Zmth", "Mathematical notation", NULL, 995},
	{"Zsye", "Symbols (Emoji variant)", NULL, 993},
	{"Zsym", "Symbols", NULL, 996},
	{"Zxxx", "Code for unwritten documents", NULL, 997},
	{"Zyyy", "Code for undetermined script", "Common", 998},
	{"Zzzz", "Code for uncoded script", "Unknown", 999},
};

