package org.coolreader.plugins;

public class OnlineStoreBookInfo implements AsyncResponse {
	public String login;
	public boolean isLoggedIn;
	public double accountBalance;
	public boolean isPurchased;
	public OnlineStoreBook book;
}
