
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int cu_lang_orth_chars[] = {
	0xf0f0ffff, 0x0401, 0x0402,	// range
	0xf0f0ffff, 0x0405, 0x0406,	// range
	0x0408,
	0x040b,
	0x040d,
	0xf0f0ffff, 0x040f, 0x0418,	// range
	0xf0f0ffff, 0x041a, 0x042c,	// range
	0xf0f0ffff, 0x042e, 0x044c,	// range
	0xf0f0ffff, 0x044e, 0x0450,	// range
	0x0452,
	0xf0f0ffff, 0x0455, 0x0456,	// range
	0x0458,
	0x045b,
	0x045d,
	0xf0f0ffff, 0x045f, 0x0479,	// range
};
#define CU_LANG_ORTH_SZ	31
