
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int ml_lang_orth_chars[] = {
	0xf0f0ffff, 0x0d02, 0x0d03,	// range
	0xf0f0ffff, 0x0d05, 0x0d0c,	// range
	0xf0f0ffff, 0x0d0e, 0x0d10,	// range
	0xf0f0ffff, 0x0d12, 0x0d28,	// range
	0xf0f0ffff, 0x0d2a, 0x0d39,	// range
	0xf0f0ffff, 0x0d3e, 0x0d43,	// range
	0xf0f0ffff, 0x0d46, 0x0d48,	// range
	0xf0f0ffff, 0x0d4a, 0x0d4d,	// range
	0x0d57,
	0xf0f0ffff, 0x0d60, 0x0d61,	// range
};
#define ML_LANG_ORTH_SZ	28
