// ISO-639-2 Language registry.
// License: Public Domain.
// This file is autogenerated.
// Based on data from: https://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt
// file: ISO-639-2_utf-8.txt
// url: https://www.loc.gov/standards/iso639-2/ISO-639-2_utf-8.txt

#include "iso-639-2_data.h"

#include <stddef.h>

const struct iso639_2_rec iso639_2_data[] = {
	{"abk", NULL, "ab", "Abkhazian"},
	{"ace", NULL, NULL, "Achinese"},
	{"ach", NULL, NULL, "Acoli"},
	{"ada", NULL, NULL, "Adangme"},
	{"ady", NULL, NULL, "Adyghe; Adygei"},
	{"afa", NULL, NULL, "Afro-Asiatic languages"},
	{"afh", NULL, NULL, "Afrihili"},
	{"afr", NULL, "af", "Afrikaans"},
	{"ain", NULL, NULL, "Ainu"},
	{"aka", NULL, "ak", "Akan"},
	{"akk", NULL, NULL, "Akkadian"},
	{"alb", "sqi", "sq", "Albanian"},
	{"ale", NULL, NULL, "Aleut"},
	{"alg", NULL, NULL, "Algonquian languages"},
	{"alt", NULL, NULL, "Southern Altai"},
	{"amh", NULL, "am", "Amharic"},
	{"ang", NULL, NULL, "English, Old (ca.450-1100)"},
	{"anp", NULL, NULL, "Angika"},
	{"apa", NULL, NULL, "Apache languages"},
	{"ara", NULL, "ar", "Arabic"},
	{"arc", NULL, NULL, "Official Aramaic (700-300 BCE); Imperial Aramaic (700-300 BCE)"},
	{"arg", NULL, "an", "Aragonese"},
	{"arm", "hye", "hy", "Armenian"},
	{"arn", NULL, NULL, "Mapudungun; Mapuche"},
	{"arp", NULL, NULL, "Arapaho"},
	{"art", NULL, NULL, "Artificial languages"},
	{"arw", NULL, NULL, "Arawak"},
	{"asm", NULL, "as", "Assamese"},
	{"ast", NULL, NULL, "Asturian; Bable; Leonese; Asturleonese"},
	{"ath", NULL, NULL, "Athapascan languages"},
	{"aus", NULL, NULL, "Australian languages"},
	{"ava", NULL, "av", "Avaric"},
	{"ave", NULL, "ae", "Avestan"},
	{"awa", NULL, NULL, "Awadhi"},
	{"aym", NULL, "ay", "Aymara"},
	{"aze", NULL, "az", "Azerbaijani"},
	{"bad", NULL, NULL, "Banda languages"},
	{"bai", NULL, NULL, "Bamileke languages"},
	{"bak", NULL, "ba", "Bashkir"},
	{"bal", NULL, NULL, "Baluchi"},
	{"bam", NULL, "bm", "Bambara"},
	{"ban", NULL, NULL, "Balinese"},
	{"baq", "eus", "eu", "Basque"},
	{"bas", NULL, NULL, "Basa"},
	{"bat", NULL, NULL, "Baltic languages"},
	{"bej", NULL, NULL, "Beja; Bedawiyet"},
	{"bel", NULL, "be", "Belarusian"},
	{"bem", NULL, NULL, "Bemba"},
	{"ben", NULL, "bn", "Bengali"},
	{"ber", NULL, NULL, "Berber languages"},
	{"bho", NULL, NULL, "Bhojpuri"},
	{"bih", NULL, "bh", "Bihari languages"},
	{"bik", NULL, NULL, "Bikol"},
	{"bin", NULL, NULL, "Bini; Edo"},
	{"bis", NULL, "bi", "Bislama"},
	{"bla", NULL, NULL, "Siksika"},
	{"bnt", NULL, NULL, "Bantu languages"},
	{"bos", NULL, "bs", "Bosnian"},
	{"bra", NULL, NULL, "Braj"},
	{"bre", NULL, "br", "Breton"},
	{"btk", NULL, NULL, "Batak languages"},
	{"bua", NULL, NULL, "Buriat"},
	{"bug", NULL, NULL, "Buginese"},
	{"bul", NULL, "bg", "Bulgarian"},
	{"bur", "mya", "my", "Burmese"},
	{"byn", NULL, NULL, "Blin; Bilin"},
	{"cad", NULL, NULL, "Caddo"},
	{"cai", NULL, NULL, "Central American Indian languages"},
	{"car", NULL, NULL, "Galibi Carib"},
	{"cat", NULL, "ca", "Catalan; Valencian"},
	{"cau", NULL, NULL, "Caucasian languages"},
	{"ceb", NULL, NULL, "Cebuano"},
	{"cel", NULL, NULL, "Celtic languages"},
	{"cha", NULL, "ch", "Chamorro"},
	{"chb", NULL, NULL, "Chibcha"},
	{"che", NULL, "ce", "Chechen"},
	{"chg", NULL, NULL, "Chagatai"},
	{"chi", "zho", "zh", "Chinese"},
	{"chk", NULL, NULL, "Chuukese"},
	{"chm", NULL, NULL, "Mari"},
	{"chn", NULL, NULL, "Chinook jargon"},
	{"cho", NULL, NULL, "Choctaw"},
	{"chp", NULL, NULL, "Chipewyan; Dene Suline"},
	{"chr", NULL, NULL, "Cherokee"},
	{"chu", NULL, "cu", "Church Slavic; Old Slavonic; Church Slavonic; Old Bulgarian; Old Church Slavonic"},
	{"chv", NULL, "cv", "Chuvash"},
	{"chy", NULL, NULL, "Cheyenne"},
	{"cmc", NULL, NULL, "Chamic languages"},
	{"cnr", NULL, NULL, "Montenegrin"},
	{"cop", NULL, NULL, "Coptic"},
	{"cor", NULL, "kw", "Cornish"},
	{"cos", NULL, "co", "Corsican"},
	{"cpe", NULL, NULL, "Creoles and pidgins, English based"},
	{"cpf", NULL, NULL, "Creoles and pidgins, French-based"},
	{"cpp", NULL, NULL, "Creoles and pidgins, Portuguese-based"},
	{"cre", NULL, "cr", "Cree"},
	{"crh", NULL, NULL, "Crimean Tatar; Crimean Turkish"},
	{"crp", NULL, NULL, "Creoles and pidgins"},
	{"csb", NULL, NULL, "Kashubian"},
	{"cus", NULL, NULL, "Cushitic languages"},
	{"cze", "ces", "cs", "Czech"},
	{"dak", NULL, NULL, "Dakota"},
	{"dan", NULL, "da", "Danish"},
	{"dar", NULL, NULL, "Dargwa"},
	{"day", NULL, NULL, "Land Dayak languages"},
	{"del", NULL, NULL, "Delaware"},
	{"den", NULL, NULL, "Slave (Athapascan)"},
	{"dgr", NULL, NULL, "Dogrib"},
	{"din", NULL, NULL, "Dinka"},
	{"div", NULL, "dv", "Divehi; Dhivehi; Maldivian"},
	{"doi", NULL, NULL, "Dogri"},
	{"dra", NULL, NULL, "Dravidian languages"},
	{"dsb", NULL, NULL, "Lower Sorbian"},
	{"dua", NULL, NULL, "Duala"},
	{"dum", NULL, NULL, "Dutch, Middle (ca.1050-1350)"},
	{"dut", "nld", "nl", "Dutch; Flemish"},
	{"dyu", NULL, NULL, "Dyula"},
	{"dzo", NULL, "dz", "Dzongkha"},
	{"efi", NULL, NULL, "Efik"},
	{"egy", NULL, NULL, "Egyptian (Ancient)"},
	{"eka", NULL, NULL, "Ekajuk"},
	{"elx", NULL, NULL, "Elamite"},
	{"eng", NULL, "en", "English"},
	{"enm", NULL, NULL, "English, Middle (1100-1500)"},
	{"epo", NULL, "eo", "Esperanto"},
	{"est", NULL, "et", "Estonian"},
	{"ewe", NULL, "ee", "Ewe"},
	{"ewo", NULL, NULL, "Ewondo"},
	{"fan", NULL, NULL, "Fang"},
	{"fao", NULL, "fo", "Faroese"},
	{"fat", NULL, NULL, "Fanti"},
	{"fij", NULL, "fj", "Fijian"},
	{"fil", NULL, NULL, "Filipino; Pilipino"},
	{"fin", NULL, "fi", "Finnish"},
	{"fiu", NULL, NULL, "Finno-Ugrian languages"},
	{"fon", NULL, NULL, "Fon"},
	{"fre", "fra", "fr", "French"},
	{"frm", NULL, NULL, "French, Middle (ca.1400-1600)"},
	{"fro", NULL, NULL, "French, Old (842-ca.1400)"},
	{"frr", NULL, NULL, "Northern Frisian"},
	{"frs", NULL, NULL, "Eastern Frisian"},
	{"fry", NULL, "fy", "Western Frisian"},
	{"ful", NULL, "ff", "Fulah"},
	{"fur", NULL, NULL, "Friulian"},
	{"gaa", NULL, NULL, "Ga"},
	{"gay", NULL, NULL, "Gayo"},
	{"gba", NULL, NULL, "Gbaya"},
	{"gem", NULL, NULL, "Germanic languages"},
	{"geo", "kat", "ka", "Georgian"},
	{"ger", "deu", "de", "German"},
	{"gez", NULL, NULL, "Geez"},
	{"gil", NULL, NULL, "Gilbertese"},
	{"gla", NULL, "gd", "Gaelic; Scottish Gaelic"},
	{"gle", NULL, "ga", "Irish"},
	{"glg", NULL, "gl", "Galician"},
	{"glv", NULL, "gv", "Manx"},
	{"gmh", NULL, NULL, "German, Middle High (ca.1050-1500)"},
	{"goh", NULL, NULL, "German, Old High (ca.750-1050)"},
	{"gon", NULL, NULL, "Gondi"},
	{"gor", NULL, NULL, "Gorontalo"},
	{"got", NULL, NULL, "Gothic"},
	{"grb", NULL, NULL, "Grebo"},
	{"grc", NULL, NULL, "Greek, Ancient (to 1453)"},
	{"gre", "ell", "el", "Greek, Modern (1453-)"},
	{"grn", NULL, "gn", "Guarani"},
	{"gsw", NULL, NULL, "Swiss German; Alemannic; Alsatian"},
	{"guj", NULL, "gu", "Gujarati"},
	{"gwi", NULL, NULL, "Gwich'in"},
	{"hai", NULL, NULL, "Haida"},
	{"hat", NULL, "ht", "Haitian; Haitian Creole"},
	{"hau", NULL, "ha", "Hausa"},
	{"haw", NULL, NULL, "Hawaiian"},
	{"heb", NULL, "he", "Hebrew"},
	{"her", NULL, "hz", "Herero"},
	{"hil", NULL, NULL, "Hiligaynon"},
	{"him", NULL, NULL, "Himachali languages; Western Pahari languages"},
	{"hin", NULL, "hi", "Hindi"},
	{"hit", NULL, NULL, "Hittite"},
	{"hmn", NULL, NULL, "Hmong; Mong"},
	{"hmo", NULL, "ho", "Hiri Motu"},
	{"hrv", NULL, "hr", "Croatian"},
	{"hsb", NULL, NULL, "Upper Sorbian"},
	{"hun", NULL, "hu", "Hungarian"},
	{"hup", NULL, NULL, "Hupa"},
	{"iba", NULL, NULL, "Iban"},
	{"ibo", NULL, "ig", "Igbo"},
	{"ice", "isl", "is", "Icelandic"},
	{"ido", NULL, "io", "Ido"},
	{"iii", NULL, "ii", "Sichuan Yi; Nuosu"},
	{"ijo", NULL, NULL, "Ijo languages"},
	{"iku", NULL, "iu", "Inuktitut"},
	{"ile", NULL, "ie", "Interlingue; Occidental"},
	{"ilo", NULL, NULL, "Iloko"},
	{"ina", NULL, "ia", "Interlingua (International Auxiliary Language Association)"},
	{"inc", NULL, NULL, "Indic languages"},
	{"ind", NULL, "id", "Indonesian"},
	{"ine", NULL, NULL, "Indo-European languages"},
	{"inh", NULL, NULL, "Ingush"},
	{"ipk", NULL, "ik", "Inupiaq"},
	{"ira", NULL, NULL, "Iranian languages"},
	{"iro", NULL, NULL, "Iroquoian languages"},
	{"ita", NULL, "it", "Italian"},
	{"jav", NULL, "jv", "Javanese"},
	{"jbo", NULL, NULL, "Lojban"},
	{"jpn", NULL, "ja", "Japanese"},
	{"jpr", NULL, NULL, "Judeo-Persian"},
	{"jrb", NULL, NULL, "Judeo-Arabic"},
	{"kaa", NULL, NULL, "Kara-Kalpak"},
	{"kab", NULL, NULL, "Kabyle"},
	{"kac", NULL, NULL, "Kachin; Jingpho"},
	{"kal", NULL, "kl", "Kalaallisut; Greenlandic"},
	{"kam", NULL, NULL, "Kamba"},
	{"kan", NULL, "kn", "Kannada"},
	{"kar", NULL, NULL, "Karen languages"},
	{"kas", NULL, "ks", "Kashmiri"},
	{"kau", NULL, "kr", "Kanuri"},
	{"kaw", NULL, NULL, "Kawi"},
	{"kaz", NULL, "kk", "Kazakh"},
	{"kbd", NULL, NULL, "Kabardian"},
	{"kha", NULL, NULL, "Khasi"},
	{"khi", NULL, NULL, "Khoisan languages"},
	{"khm", NULL, "km", "Central Khmer"},
	{"kho", NULL, NULL, "Khotanese; Sakan"},
	{"kik", NULL, "ki", "Kikuyu; Gikuyu"},
	{"kin", NULL, "rw", "Kinyarwanda"},
	{"kir", NULL, "ky", "Kirghiz; Kyrgyz"},
	{"kmb", NULL, NULL, "Kimbundu"},
	{"kok", NULL, NULL, "Konkani"},
	{"kom", NULL, "kv", "Komi"},
	{"kon", NULL, "kg", "Kongo"},
	{"kor", NULL, "ko", "Korean"},
	{"kos", NULL, NULL, "Kosraean"},
	{"kpe", NULL, NULL, "Kpelle"},
	{"krc", NULL, NULL, "Karachay-Balkar"},
	{"krl", NULL, NULL, "Karelian"},
	{"kro", NULL, NULL, "Kru languages"},
	{"kru", NULL, NULL, "Kurukh"},
	{"kua", NULL, "kj", "Kuanyama; Kwanyama"},
	{"kum", NULL, NULL, "Kumyk"},
	{"kur", NULL, "ku", "Kurdish"},
	{"kut", NULL, NULL, "Kutenai"},
	{"lad", NULL, NULL, "Ladino"},
	{"lah", NULL, NULL, "Lahnda"},
	{"lam", NULL, NULL, "Lamba"},
	{"lao", NULL, "lo", "Lao"},
	{"lat", NULL, "la", "Latin"},
	{"lav", NULL, "lv", "Latvian"},
	{"lez", NULL, NULL, "Lezghian"},
	{"lim", NULL, "li", "Limburgan; Limburger; Limburgish"},
	{"lin", NULL, "ln", "Lingala"},
	{"lit", NULL, "lt", "Lithuanian"},
	{"lol", NULL, NULL, "Mongo"},
	{"loz", NULL, NULL, "Lozi"},
	{"ltz", NULL, "lb", "Luxembourgish; Letzeburgesch"},
	{"lua", NULL, NULL, "Luba-Lulua"},
	{"lub", NULL, "lu", "Luba-Katanga"},
	{"lug", NULL, "lg", "Ganda"},
	{"lui", NULL, NULL, "Luiseno"},
	{"lun", NULL, NULL, "Lunda"},
	{"luo", NULL, NULL, "Luo (Kenya and Tanzania)"},
	{"lus", NULL, NULL, "Lushai"},
	{"mac", "mkd", "mk", "Macedonian"},
	{"mad", NULL, NULL, "Madurese"},
	{"mag", NULL, NULL, "Magahi"},
	{"mah", NULL, "mh", "Marshallese"},
	{"mai", NULL, NULL, "Maithili"},
	{"mak", NULL, NULL, "Makasar"},
	{"mal", NULL, "ml", "Malayalam"},
	{"man", NULL, NULL, "Mandingo"},
	{"mao", "mri", "mi", "Maori"},
	{"map", NULL, NULL, "Austronesian languages"},
	{"mar", NULL, "mr", "Marathi"},
	{"mas", NULL, NULL, "Masai"},
	{"may", "msa", "ms", "Malay"},
	{"mdf", NULL, NULL, "Moksha"},
	{"mdr", NULL, NULL, "Mandar"},
	{"men", NULL, NULL, "Mende"},
	{"mga", NULL, NULL, "Irish, Middle (900-1200)"},
	{"mic", NULL, NULL, "Mi'kmaq; Micmac"},
	{"min", NULL, NULL, "Minangkabau"},
	{"mis", NULL, NULL, "Uncoded languages"},
	{"mkh", NULL, NULL, "Mon-Khmer languages"},
	{"mlg", NULL, "mg", "Malagasy"},
	{"mlt", NULL, "mt", "Maltese"},
	{"mnc", NULL, NULL, "Manchu"},
	{"mni", NULL, NULL, "Manipuri"},
	{"mno", NULL, NULL, "Manobo languages"},
	{"moh", NULL, NULL, "Mohawk"},
	{"mon", NULL, "mn", "Mongolian"},
	{"mos", NULL, NULL, "Mossi"},
	{"mul", NULL, NULL, "Multiple languages"},
	{"mun", NULL, NULL, "Munda languages"},
	{"mus", NULL, NULL, "Creek"},
	{"mwl", NULL, NULL, "Mirandese"},
	{"mwr", NULL, NULL, "Marwari"},
	{"myn", NULL, NULL, "Mayan languages"},
	{"myv", NULL, NULL, "Erzya"},
	{"nah", NULL, NULL, "Nahuatl languages"},
	{"nai", NULL, NULL, "North American Indian languages"},
	{"nap", NULL, NULL, "Neapolitan"},
	{"nau", NULL, "na", "Nauru"},
	{"nav", NULL, "nv", "Navajo; Navaho"},
	{"nbl", NULL, "nr", "Ndebele, South; South Ndebele"},
	{"nde", NULL, "nd", "Ndebele, North; North Ndebele"},
	{"ndo", NULL, "ng", "Ndonga"},
	{"nds", NULL, NULL, "Low German; Low Saxon; German, Low; Saxon, Low"},
	{"nep", NULL, "ne", "Nepali"},
	{"new", NULL, NULL, "Nepal Bhasa; Newari"},
	{"nia", NULL, NULL, "Nias"},
	{"nic", NULL, NULL, "Niger-Kordofanian languages"},
	{"niu", NULL, NULL, "Niuean"},
	{"nno", NULL, "nn", "Norwegian Nynorsk; Nynorsk, Norwegian"},
	{"nob", NULL, "nb", "Bokmål, Norwegian; Norwegian Bokmål"},
	{"nog", NULL, NULL, "Nogai"},
	{"non", NULL, NULL, "Norse, Old"},
	{"nor", NULL, "no", "Norwegian"},
	{"nqo", NULL, NULL, "N'Ko"},
	{"nso", NULL, NULL, "Pedi; Sepedi; Northern Sotho"},
	{"nub", NULL, NULL, "Nubian languages"},
	{"nwc", NULL, NULL, "Classical Newari; Old Newari; Classical Nepal Bhasa"},
	{"nya", NULL, "ny", "Chichewa; Chewa; Nyanja"},
	{"nym", NULL, NULL, "Nyamwezi"},
	{"nyn", NULL, NULL, "Nyankole"},
	{"nyo", NULL, NULL, "Nyoro"},
	{"nzi", NULL, NULL, "Nzima"},
	{"oci", NULL, "oc", "Occitan (post 1500)"},
	{"oji", NULL, "oj", "Ojibwa"},
	{"ori", NULL, "or", "Oriya"},
	{"orm", NULL, "om", "Oromo"},
	{"osa", NULL, NULL, "Osage"},
	{"oss", NULL, "os", "Ossetian; Ossetic"},
	{"ota", NULL, NULL, "Turkish, Ottoman (1500-1928)"},
	{"oto", NULL, NULL, "Otomian languages"},
	{"paa", NULL, NULL, "Papuan languages"},
	{"pag", NULL, NULL, "Pangasinan"},
	{"pal", NULL, NULL, "Pahlavi"},
	{"pam", NULL, NULL, "Pampanga; Kapampangan"},
	{"pan", NULL, "pa", "Panjabi; Punjabi"},
	{"pap", NULL, NULL, "Papiamento"},
	{"pau", NULL, NULL, "Palauan"},
	{"peo", NULL, NULL, "Persian, Old (ca.600-400 B.C.)"},
	{"per", "fas", "fa", "Persian"},
	{"phi", NULL, NULL, "Philippine languages"},
	{"phn", NULL, NULL, "Phoenician"},
	{"pli", NULL, "pi", "Pali"},
	{"pol", NULL, "pl", "Polish"},
	{"pon", NULL, NULL, "Pohnpeian"},
	{"por", NULL, "pt", "Portuguese"},
	{"pra", NULL, NULL, "Prakrit languages"},
	{"pro", NULL, NULL, "Provençal, Old (to 1500); Occitan, Old (to 1500)"},
	{"pus", NULL, "ps", "Pushto; Pashto"},
	{"que", NULL, "qu", "Quechua"},
	{"raj", NULL, NULL, "Rajasthani"},
	{"rap", NULL, NULL, "Rapanui"},
	{"rar", NULL, NULL, "Rarotongan; Cook Islands Maori"},
	{"roa", NULL, NULL, "Romance languages"},
	{"roh", NULL, "rm", "Romansh"},
	{"rom", NULL, NULL, "Romany"},
	{"rum", "ron", "ro", "Romanian; Moldavian; Moldovan"},
	{"run", NULL, "rn", "Rundi"},
	{"rup", NULL, NULL, "Aromanian; Arumanian; Macedo-Romanian"},
	{"rus", NULL, "ru", "Russian"},
	{"sad", NULL, NULL, "Sandawe"},
	{"sag", NULL, "sg", "Sango"},
	{"sah", NULL, NULL, "Yakut"},
	{"sai", NULL, NULL, "South American Indian languages"},
	{"sal", NULL, NULL, "Salishan languages"},
	{"sam", NULL, NULL, "Samaritan Aramaic"},
	{"san", NULL, "sa", "Sanskrit"},
	{"sas", NULL, NULL, "Sasak"},
	{"sat", NULL, NULL, "Santali"},
	{"scn", NULL, NULL, "Sicilian"},
	{"sco", NULL, NULL, "Scots"},
	{"sel", NULL, NULL, "Selkup"},
	{"sem", NULL, NULL, "Semitic languages"},
	{"sga", NULL, NULL, "Irish, Old (to 900)"},
	{"sgn", NULL, NULL, "Sign Languages"},
	{"shn", NULL, NULL, "Shan"},
	{"sid", NULL, NULL, "Sidamo"},
	{"sin", NULL, "si", "Sinhala; Sinhalese"},
	{"sio", NULL, NULL, "Siouan languages"},
	{"sit", NULL, NULL, "Sino-Tibetan languages"},
	{"sla", NULL, NULL, "Slavic languages"},
	{"slo", "slk", "sk", "Slovak"},
	{"slv", NULL, "sl", "Slovenian"},
	{"sma", NULL, NULL, "Southern Sami"},
	{"sme", NULL, "se", "Northern Sami"},
	{"smi", NULL, NULL, "Sami languages"},
	{"smj", NULL, NULL, "Lule Sami"},
	{"smn", NULL, NULL, "Inari Sami"},
	{"smo", NULL, "sm", "Samoan"},
	{"sms", NULL, NULL, "Skolt Sami"},
	{"sna", NULL, "sn", "Shona"},
	{"snd", NULL, "sd", "Sindhi"},
	{"snk", NULL, NULL, "Soninke"},
	{"sog", NULL, NULL, "Sogdian"},
	{"som", NULL, "so", "Somali"},
	{"son", NULL, NULL, "Songhai languages"},
	{"sot", NULL, "st", "Sotho, Southern"},
	{"spa", NULL, "es", "Spanish; Castilian"},
	{"srd", NULL, "sc", "Sardinian"},
	{"srn", NULL, NULL, "Sranan Tongo"},
	{"srp", NULL, "sr", "Serbian"},
	{"srr", NULL, NULL, "Serer"},
	{"ssa", NULL, NULL, "Nilo-Saharan languages"},
	{"ssw", NULL, "ss", "Swati"},
	{"suk", NULL, NULL, "Sukuma"},
	{"sun", NULL, "su", "Sundanese"},
	{"sus", NULL, NULL, "Susu"},
	{"sux", NULL, NULL, "Sumerian"},
	{"swa", NULL, "sw", "Swahili"},
	{"swe", NULL, "sv", "Swedish"},
	{"syc", NULL, NULL, "Classical Syriac"},
	{"syr", NULL, NULL, "Syriac"},
	{"tah", NULL, "ty", "Tahitian"},
	{"tai", NULL, NULL, "Tai languages"},
	{"tam", NULL, "ta", "Tamil"},
	{"tat", NULL, "tt", "Tatar"},
	{"tel", NULL, "te", "Telugu"},
	{"tem", NULL, NULL, "Timne"},
	{"ter", NULL, NULL, "Tereno"},
	{"tet", NULL, NULL, "Tetum"},
	{"tgk", NULL, "tg", "Tajik"},
	{"tgl", NULL, "tl", "Tagalog"},
	{"tha", NULL, "th", "Thai"},
	{"tib", "bod", "bo", "Tibetan"},
	{"tig", NULL, NULL, "Tigre"},
	{"tir", NULL, "ti", "Tigrinya"},
	{"tiv", NULL, NULL, "Tiv"},
	{"tkl", NULL, NULL, "Tokelau"},
	{"tlh", NULL, NULL, "Klingon; tlhIngan-Hol"},
	{"tli", NULL, NULL, "Tlingit"},
	{"tmh", NULL, NULL, "Tamashek"},
	{"tog", NULL, NULL, "Tonga (Nyasa)"},
	{"ton", NULL, "to", "Tonga (Tonga Islands)"},
	{"tpi", NULL, NULL, "Tok Pisin"},
	{"tsi", NULL, NULL, "Tsimshian"},
	{"tsn", NULL, "tn", "Tswana"},
	{"tso", NULL, "ts", "Tsonga"},
	{"tuk", NULL, "tk", "Turkmen"},
	{"tum", NULL, NULL, "Tumbuka"},
	{"tup", NULL, NULL, "Tupi languages"},
	{"tur", NULL, "tr", "Turkish"},
	{"tut", NULL, NULL, "Altaic languages"},
	{"tvl", NULL, NULL, "Tuvalu"},
	{"twi", NULL, "tw", "Twi"},
	{"tyv", NULL, NULL, "Tuvinian"},
	{"udm", NULL, NULL, "Udmurt"},
	{"uga", NULL, NULL, "Ugaritic"},
	{"uig", NULL, "ug", "Uighur; Uyghur"},
	{"ukr", NULL, "uk", "Ukrainian"},
	{"umb", NULL, NULL, "Umbundu"},
	{"und", NULL, NULL, "Undetermined"},
	{"urd", NULL, "ur", "Urdu"},
	{"uzb", NULL, "uz", "Uzbek"},
	{"vai", NULL, NULL, "Vai"},
	{"ven", NULL, "ve", "Venda"},
	{"vie", NULL, "vi", "Vietnamese"},
	{"vol", NULL, "vo", "Volapük"},
	{"vot", NULL, NULL, "Votic"},
	{"wak", NULL, NULL, "Wakashan languages"},
	{"wal", NULL, NULL, "Wolaitta; Wolaytta"},
	{"war", NULL, NULL, "Waray"},
	{"was", NULL, NULL, "Washo"},
	{"wel", "cym", "cy", "Welsh"},
	{"wen", NULL, NULL, "Sorbian languages"},
	{"wln", NULL, "wa", "Walloon"},
	{"wol", NULL, "wo", "Wolof"},
	{"xal", NULL, NULL, "Kalmyk; Oirat"},
	{"xho", NULL, "xh", "Xhosa"},
	{"yao", NULL, NULL, "Yao"},
	{"yap", NULL, NULL, "Yapese"},
	{"yid", NULL, "yi", "Yiddish"},
	{"yor", NULL, "yo", "Yoruba"},
	{"ypk", NULL, NULL, "Yupik languages"},
	{"zap", NULL, NULL, "Zapotec"},
	{"zbl", NULL, NULL, "Blissymbols; Blissymbolics; Bliss"},
	{"zen", NULL, NULL, "Zenaga"},
	{"zgh", NULL, NULL, "Standard Moroccan Tamazight"},
	{"zha", NULL, "za", "Zhuang; Chuang"},
	{"znd", NULL, NULL, "Zande languages"},
	{"zul", NULL, "zu", "Zulu"},
	{"zun", NULL, NULL, "Zuni"},
	{"zxx", NULL, NULL, "No linguistic content; Not applicable"},
	{"zza", NULL, NULL, "Zaza; Dimili; Dimli; Kirdki; Kirmanjki; Zazaki"},
};

