
// This file is autogenerated from fc-lang database.
// https://www.freedesktop.org/wiki/Software/fontconfig/
// https://gitlab.freedesktop.org/fontconfig/fontconfig/tree/master/fc-lang
// by convert utility from https://github.com/virxkane/freetype_textdraw

const unsigned int gu_lang_orth_chars[] = {
	0xf0f0ffff, 0x0a81, 0x0a83,	// range
	0xf0f0ffff, 0x0a85, 0x0a8b,	// range
	0x0a8d,
	0xf0f0ffff, 0x0a8f, 0x0a91,	// range
	0xf0f0ffff, 0x0a93, 0x0aa8,	// range
	0xf0f0ffff, 0x0aaa, 0x0ab0,	// range
	0xf0f0ffff, 0x0ab2, 0x0ab3,	// range
	0xf0f0ffff, 0x0ab5, 0x0ab9,	// range
	0xf0f0ffff, 0x0abc, 0x0ac5,	// range
	0xf0f0ffff, 0x0ac7, 0x0ac9,	// range
	0xf0f0ffff, 0x0acb, 0x0acd,	// range
	0x0ad0,
	0x0ae0,
};
#define GU_LANG_ORTH_SZ	33
